<?php 
require_once __DIR__.'/../libs/RemoteIOModule.inc';
class BaseClientModule extends RemoteIOModule {
	public function Create(){
		parent::Create();
		$this->RegisterPropertyBoolean('EnableConfig', true);
		$this->RegisterPropertyInteger('CreateCategoryID', 0);
		$this->RegisterPropertyBoolean('EnableControlButtons', true);
		$this->RegisterPropertyBoolean('AutoCreateLinks', false);
		$this->RegisterPropertyBoolean('RemoveUnusedProfiles', true);
		$this->RegisterPropertyInteger('WebFrontInstance', 0);
 		$this->RegisterPropertyString('LastDisableStates','');
		$this->SetBuffer('ConnectionState', CLIENT_DISCONNECTED);
 	}
	public function ApplyChanges(){
		$status=$this->GetStatus();
		if($status==102||$status==300){
			if($this->ReadPropertyInteger('CreateCategoryID')<10000)
				$this->SetStatus($status=300);
			elseif($status!=102)$this->SetStatus($status=102);
		}
		if($status==102 || $status==301){
			if($webID=$this->ReadPropertyInteger('WebFrontInstance')){
				$i=@IPS_GetInstance($webID);
				if($i['ModuleInfo']['ModuleID']!='{3565B1F2-8F7B-4311-A4B6-1BF1D868F39E}'){
					if($status!=301)$this->SetStatus($status=301);
				}elseif($status!=102)$this->SetStatus($status=102);
			}elseif($status!=102)$this->SetStatus($status=102);
		}
		if($status!=102)$this->BlockApply();
		parent::ApplyChanges();
	}
	// -----------------------------------------
	// -  Form Maintain
	// -----------------------------------------
	protected function GetFormConfigItems($FormVersion){
		if($this->ReadPropertyBoolean('EnableConfig')){
			$items=parent::GetFormConfigItems($FormVersion);
			$items[]=["type"=> "SelectCategory","name"=> "CreateCategoryID","caption"=> "Create links in category"];
			$items[]=["type"=> "CheckBox","name"=> "EnableControlButtons","caption"=> "Show control buttons"];
			$items[]=["type"=> "CheckBox", "name"=> "AutoCreateLinks", "caption"=> "Auto create links"];
			$items[]=["type"=> "CheckBox", "name"=> "RemoveUnusedProfiles", "caption"=> "Remove unused Profiles"];
			$items[]=["type"=> "SelectInstance", "name"=> "WebFrontInstance", "caption"=> "Select WebFrontend to show errors"];
		} else $items=null;
		return $items;
	}
	protected function GetFormButtons($FormVersion){
		$items=parent::GetFormButtons($FormVersion);
		$disabled=$this->IsDisabled();
		$values=$this->GetFormListValues($FormVersion);
		$hasValues=count($values)>0;
		$hasLinks=$hasValues && count(array_filter($values,function($s){return $s[SHARE_LOCALID]>0;}))>0;
		if($this->ReadPropertyBoolean('EnableControlButtons')){
			$caption	= $FormVersion<5?'label':'caption';
			$reconnect="IPS_RequestAction($this->InstanceID,'RE_CONNECT',true);";
			$confirm=$this->Translate("Reconnect Client ?");
			$items[]=["type"=>"Button",$caption=>"Reconnect Client","onClick"=>$reconnect, "confirm"=>$confirm];
			if($disabled){
				$info=$this->Translate("Client is not allowed. Check your Server settins!");
				$items[]=["type"=>"Button",$caption=>"Client not Allowed","onClick"=>"echo '$info';"];
			}
			$wait=$this->Translate("Please wait a moment, the request has been sent");
			if($FormVersion >4 && $FormVersion!=FOMR_CONSOLE){
				$errSelected = $this->Translate("Selected object");
				$errRoot = $this->Translate("is not a root object");
				if($hasLinks){
					$confirm=$this->Translate("Do you really want to disconnect the remote connection?");
					$err = $this->Translate("Please select a root object to disconnect");
					$errConn = $this->Translate("is not connected");
					$delete="if(empty(\$ShareList) || \$ShareList['".SHARE_OBJ_ID."']==0)echo '$err';
					elseif(\$ShareList['".SHARE_PARENT."']>0)echo '$errSelected \"'.\$ShareList['".SHARE_NAME."'].'\" $errRoot';
					elseif(\$ShareList['".SHARE_LOCALID."']==0)echo '$errSelected \"'.\$ShareList['".SHARE_NAME."'].'\" $errConn';
					else if(IPS_RequestAction($this->InstanceID, 'OBJECT_DELETE', \$ShareList['".SHARE_OBJ_ID."']))echo '$wait';";
					$items[]=["type"=>"Button",$caption=>"Disconnect object","onClick"=>$delete, "confirm"=>$confirm];
				}
				if(!$disabled && $hasValues){
					$confirm=$this->Translate("Do you want to connect remotely?");
					$err = $this->Translate("Please select a root object to connect to");
					$errConn = $this->Translate("is already connected");
					$create="if(empty(\$ShareList) || \$ShareList['".SHARE_OBJ_ID."']==0)echo '$err';
					elseif(\$ShareList['".SHARE_PARENT."']>0)echo '$errSelected \"'.\$ShareList['".SHARE_NAME."'].'\" $errRoot';
					elseif(\$ShareList['".SHARE_LOCALID."']>0)echo '$errSelected \"'.\$ShareList['".SHARE_NAME."'].'\" $errConn';
					else if(IPS_RequestAction($this->InstanceID, 'OBJECT_CREATE', \$ShareList['".SHARE_OBJ_ID."']))echo '$wait';";
					$items[]=["type"=>"Button",$caption=>"Connect object","onClick"=>$create,"confirm"=>$confirm];
				}
			}
			if(!$disabled && $hasValues){
				$confirm=$this->Translate("Do you want to connect all shares?");
				$create="if(IPS_RequestAction($this->InstanceID, 'OBJECT_CREATE_ALL',0))echo '$wait';";
				$items[]=["type"=>"Button",$caption=>"Connect All","onClick"=>$create,"confirm"=>$confirm];
			}
			if($hasLinks){
				$confirm=$this->Translate("Do you want to disconnect all shares?");
				$delete="if(IPS_RequestAction($this->InstanceID, 'OBJECT_DELETE_ALL',0))echo '$wait';";
				$items[]=["type"=>"Button",$caption=>"Disconnect All","onClick"=>$delete,"confirm"=>$confirm];
			}
		}
		return $items;
	}
	protected function GetFormList($FormVersion){
		$values		= $this->GetFormListValues($FormVersion);
		$caption	= $FormVersion<5?'label':'caption';
		$dirnames=$this->TranslatedDirNames();
		foreach($values as &$value){
			$value['DIR']=$dirnames[$value[SHARE_DIR]];
		}
		if($FormVersion!=5){
			$parent =[ "name"=> 'parent', $caption=> "Parent", "width"=> "40px","save"=>true];
			$id		=[ "name"=> 'id', 	  $caption=> "ID",     "width"=> "30px", "save"=>true];
		}else {
			$parent =["name"=>'parent', $caption=> "", "width"=>"0px","save"=>true,"visible"=>false];
			$id 	=["name"=>"id",	  $caption=> "", "width"=>"0px","save"=>true,"visible"=>false];
		}
		$list=[ "type"=> $FormVersion!=5?"List":"Tree","name"=> "ShareList", "caption"=> "Remote Shares","add"=> false,"delete"=> false,
		    "columns"=> [$id,$parent,
		    	[ "name"=> SHARE_OBJ_ID, $caption=> "Remote ID","width"=> "120px", "save"=>true], 
		    	[ "name"=> SHARE_NAME, $caption=> "Object name","width"=> "auto", "save"=>true], 
		    	[ "name"=> "type", $caption=> "Object type", "width"=> "120px", "save"=> false],
		    	[ "name"=> 'DIR', $caption=> "Direction", "width"=> "150px","save"=>false], 
		    	[ "name"=> SHARE_LOCALID, $caption=> "Local ID", "width"=> "120px", "save"=> true],
		    	[ "name"=> "state", $caption=> "State", "width"=> "100px", "save"=> false],
		    	[ "name"=> SHARE_TYPE, 	$caption=> "", "width"=> "0px", "save"=> true,"visible"=>false],
		    	[ "name"=> SHARE_DIR,  	$caption=> "", "width"=> "0px", "save"=> true,"visible"=>false],
			    [ "name"=> SHARE_STATE, $caption=> "", "width"=> "0px", "save"=> true,"visible"=>false],
		    	[ "name"=> SHARE_PARENT,$caption=> "", "width"=> "0px", "save"=> true,"visible"=>false],
			   	[ "name"=> SHARE_VTYPE, $caption=> "", "width"=> "0px", "save"=> true,"visible"=>false],
		    ],
			"values"=> $values
		];
		if($FormVersion==FOMR_CONSOLE||$FormVersion<5)$list['rowCount']=12;
		return [$list];
	}
	protected function GetFormStatusCodes(){
		$items=parent::GetFormStatusCodes();
		$items[]=[ "code"=>300, "icon"=> "error", "caption"=> "No Category for new Objects selected. Go To Settings and change it" ];
		$items[]=[ "code"=>301, "icon"=> "error", "caption"=> "Selected WebFront Instance is not a Web Front Configurator" ];
		$items[]=[ "code"=>302, "icon"=> "error", "caption"=> "Client not allowed on selected Server" ];
		return $items;
	}
	protected function GetFormListValues($FormVersion){
		$r=parent::GetFormListValues($FormVersion);
		if($this->IsDisabled()){
			$disabled=$this->TranslatedStateNames(STATE_DISABLED);
			foreach(array_keys($r) as $k){
				$r[$k]['rowColor']=COLOR_DISABLED;
				$r[$k]['state']= $disabled;
			}
		}
		return $r;
	}
	protected function GetFormActions($FormVersion){
	
		if(($FormVersion==FOMR_CONSOLE || $FormVersion < 5 )&& $this->ReadPropertyBoolean('EnableControlButtons')){
			$items=[];
 			$disabled=$this->IsDisabled();
			$values=$this->GetFormListValues($FormVersion);
			$deleteOptions=array_values(array_map(
					function($i){return ['value'=>$i[SHARE_OBJ_ID],"label"=>"({$i[SHARE_OBJ_ID]}) ".$i[SHARE_NAME]];},
					array_filter($values,function($i){return $i[SHARE_LOCALID]>0 && $i[SHARE_PARENT]==0;})
			));	
 			if(count($deleteOptions)>0){
				$err=$this->Translate('Please select object to disconnect from');
 				$items[]=["type"=>"Select","name"=>"removeID", "caption"=>"Disconnect from","options"=>$deleteOptions];
				$items[]=["type"=>"Button","caption"=>"Disconnect","onClick"=>"if(is_null(\$removeID))echo'$err';else IPS_RequestAction($this->InstanceID,'OBJECT_DELETE',\$removeID);"];
 			}
 			if(!$disabled){
				$createOptions=array_values(array_map(
						function($i){return ['value'=>$i[SHARE_OBJ_ID],"label"=>"({$i[SHARE_OBJ_ID]}) ".$i[SHARE_NAME]];},
						array_filter($values,function($i){return $i[SHARE_LOCALID]==0 && $i[SHARE_PARENT]==0;})
				));		
				if(count($createOptions)>0){
					$err=$this->Translate("Please select the object to connect to");
					$items[]=["type"=>"Select","name"=>"createID", "caption"=>"Connect to","options"=>$createOptions];
	 				$items[]=["type"=>"Button","caption"=>"Connect","onClick"=>"if(is_null(\$createID))echo'$err';else IPS_RequestAction($this->InstanceID,'OBJECT_CREATE',\$createID);"];
				}
 			}
			if(count($items)>0){
// 				if($FormVersion>=5)$items=[["type"=>"RowLayout", "items"=>$items]];
				return $items;
			}
		}
		
		
	}
	protected function IsDisabled(){
		return $this->GetBuffer('ConnectionState')!=CLIENT_VERIFIED;
	}
	
}

class RemoteIOClient extends BaseClientModule {
	protected $ConnectParentGuid = '{3CFF0FD9-E306-41DB-9B5A-9D06D38576C3}';
	protected $IsServer = false;
	
	public function RequestAction($Ident, $Value){
		if($Ident=="RE_CONNECT"){
// 			$this->SaveShareIndex([]);
// 			$this->SaveShareList([]);
 			$this->SendConnectRequest();
		}
		elseif($Ident=="OBJECT_CREATE"){
			$Requests=[];
			if ($this->CreateShare($Value,$Requests)){
				$this->DoSendObjectRequest($Requests,VOPT_DISABLED+VOPT_ICON+VOPT_PROFILE+VOPT_POSITION+VOPT_VALUE);
			}
		}
		elseif($Ident=="OBJECT_CREATE_ALL"){
			$this->ConnectAll();
		}
		elseif($Ident=='OBJECT_DELETE'){
			$this->DeleteShare($Value);
		}
		elseif($Ident=='OBJECT_DELETE_ALL'){
			$this->DisconnectAll();
		}
		elseif($Ident=='TEST'){
// 			$this->SaveReferences([],false);
// 			$this->SaveShareList([],false);
// 				$this->SendDebug(__FUNCTION__.':'.__LINE__,$Value,0);
// 			echo var_export($Value,true);
		}
		else parent::RequestAction($Ident, $Value);
	}
	
	// -----------------------------------------
	// -  Override  Basic Maintain
	// -----------------------------------------
	
	protected function GetLocaleIDforRemoteID(int $RemoteID){
// 		$found=array_values(array_filter($this->LoadShareList(false),function($s)use($RemoteID){return $s[SHARE_OBJ_ID]==$RemoteID;}));
// 		$r=empty($found[0][SHARE_LOCALID])?0:$found[0][SHARE_LOCALID];
// 		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(__FUNCTION__."($RemoteID) => $r");
// 		return $r;
		
		$index=$this->LoadShareIndex(true);
		$r =empty($index[$RemoteID][SHARE_LOCALID])?0:$index[$RemoteID][SHARE_LOCALID];
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(__FUNCTION__."($RemoteID) => $r");
		return $r;
	}
	protected function GetRemoteIDforLokaleID(int $LocaleID){
		$r=key(array_filter($this->LoadShareIndex(true),function($i)use($LocaleID){return $i[SHARE_LOCALID]==$LocaleID;}));
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(__FUNCTION__."($LocaleID) => $r");
		return (int)$r;
	}
	// -----------------------------------------
	// -  Override  Events
	// -----------------------------------------
	protected function OnInstanceReady(){
		parent::OnInstanceReady();
		$this->SendConnectRequest();
	}
	protected function OnInstanceError(){
		parent::OnInstanceError();
		$this->SetBuffer('ConnectionState',CLIENT_DISCONNECTED);
 		$this->SetDisabled(true);
	}
	protected function OnControlTimer($Mode){
		parent::OnControlTimer($Mode);
		$this->DoSendPing($this->ReadPropertyString('Identifier'),true);
	}
	protected function OnPongReceived($PongData, $Ack, $Id, $ClientIP, $ClientPort){
		parent::OnPongReceived($PongData, $Ack, $Id, $ClientIP, $ClientPort);
		$this->OnStateResponse('', $PongData, false, $ClientIP, $ClientPort, $Id);
	}
	protected function OnPingReceived($PingData, $Ack, $Id, $ClientIP, $ClientPort){
		parent::OnPingReceived($PingData, $Ack, $Id, $ClientIP, $ClientPort);
		if($Ack){
			$this->DoSendPong($this->ReadPropertyString('Identifier'),false, $ClientIP, $ClientPort,$Id);
		}
	}
	protected function OnSharesResponse($Shares, $Filtered, $Ack, $Id, $ClientIP, $ClientPort){
		parent::OnSharesResponse($Shares, $Filtered, $Ack, $Id, $ClientIP, $ClientPort);
		$shareIndexDiff = $this->GetSharesIndexDiff($Shares,$this->LoadShareIndex());
		if($this->HasDiffChanges($shareIndexDiff))$this->ApplySharesIndexDiff($shareIndexDiff);
	}
	protected function OnShareChange($Share, $ValueType, $OldValue, $NewValue){
		parent::OnShareChange($Share, $ValueType, $OldValue, $NewValue);
		if($ValueType==VOPT_PROFILE && $Share[SHARE_TYPE]==OT_VARIABLE){
			if($Share[SHARE_LOCALID]>0)
				$this->UpdateVariableActionScript($Share[SHARE_LOCALID], $Share[SHARE_DIR], $NewValue>'');
		}
	}
	protected function OnShareAdd(&$Share){
		if(empty($Share[SHARE_LOCALID]))$Share[SHARE_LOCALID]=0;
		parent::OnShareAdd($Share);
	}
	protected function OnShareDelete(array $Share){
		parent::OnShareDelete($Share);
		if($Share[SHARE_LOCALID]){
			$this->OnShareDisabled($Share);
			return $this->DeleteObject($Share[SHARE_LOCALID],true);
		}
		return true;
	}
	protected function OnShareEnabled(array $Share){
		parent::OnShareEnabled($Share);
		if($Share[SHARE_LOCALID]){
			$this->RegisterObject($Share[SHARE_LOCALID],$Share[SHARE_TYPE]);
			if($this->isValidSyncMode($Share[SHARE_DIR])){
				$this->RegisterObjectMessages($Share[SHARE_LOCALID],$Share[SHARE_TYPE]);
				if(KERNEL_VERSION>5)$this->RegisterReference($Share[SHARE_LOCALID]);
			}
			else {
				if(KERNEL_VERSION>5)$this->UnRegisterReference($Share[SHARE_LOCALID]);
				$this->UnRegisterObjectMessages($Share[SHARE_LOCALID],$Share[SHARE_TYPE]);
			}
		}
	}
	protected function OnShareDisabled(array $Share){
		parent::OnShareDisabled($Share);
		if($Share[SHARE_LOCALID]){
			$this->UnRegisterObjectMessages($Share[SHARE_LOCALID],$Share[SHARE_TYPE]);
			$this->UnRegisterObject($Share[SHARE_LOCALID],$Share[SHARE_TYPE]);
			if(KERNEL_VERSION>5)$this->UnRegisterReference($Share[SHARE_LOCALID]);
		}
	}
	
	protected function OnCommandSharesChanged($Arguments){
		parent::OnCommandSharesChanged($Arguments);
		$this->DoSendChangesRequest($this->GetShareChangesRequest());
	}
	protected function OnDefaultCommandRequest($Identifier, $Command, $Arguments, $ClientIP, $ClientPort, $Id){
		parent::OnDefaultCommandRequest($Identifier, $Command, $Arguments, $ClientIP, $ClientPort, $Id);
		if($Command==CMD_RE_CONNECT)$this->SendConnectRequest();
	}
	protected function OnConfigResponse($JSONConfig, $Ack, $ClientIP, $ClientPort, $Id){
		parent::OnConfigResponse($JSONConfig, $Ack, $ClientIP, $ClientPort, $Id);
		if($cfg=json_decode($JSONConfig,true)){
			$myCfg=json_decode(IPS_GetConfiguration($this->InstanceID),true);
// $this->Debug(json_encode($myCfg));			
			foreach($cfg as $key=>$value){
				if(isset($myCfg[$key]) && $myCfg[$key]!=$value ){
					IPS_SetProperty($this->InstanceID, $key, $value);
				}
			}
			if(IPS_HasChanges($this->InstanceID))IPS_ApplyChanges($this->InstanceID);
		}
	}
	
	protected function OnAfterLocalProfileChanged($LocaleID, $NewProfile, $Modified){
		parent::OnAfterLocalProfileChanged($LocaleID, $NewProfile, $Modified);
		$shareList	= $this->LoadShareList(true);
		$found		= array_filter($shareList,function($c)use($LocaleID){return $c[SHARE_LOCALID]==$LocaleID;});
		$index		= key($found);
		if (!is_null($index)){
			$this->UpdateVariableActionScript($LocaleID,$found[$index][SHARE_DIR],true);
		}
	}
	
	protected function OnConnectResponse($Ident, $ConnectState, $ShareChangesResponset, $Ack, $ClientIP, $ClientPort, $Id){
		parent::OnConnectResponse($Ident, $ConnectState, $ShareChangesResponset, $Ack, $ClientIP, $ClientPort, $Id);
		$this->OnStateResponse('', $ConnectState, false, $ClientIP, $ClientPort, $Id);
		if($ConnectState==CLIENT_VERIFIED){
			if($this->HasDiffChanges($ShareChangesResponset)){
				$this->ApplyShareChangesResponse($ShareChangesResponset);
			};
		}

	}
	
	protected function OnStateResponse($Ident, $ClientState, $Ack, $ClientIP, $ClientPort, $Id){
		parent::OnStateResponse($Ident, $ClientState, $Ack, $ClientIP, $ClientPort, $Id);
		$OldState=$this->GetBuffer('ConnectionState');
		$this->SetBuffer('ConnectionState',$ClientState);
		if($OldState != $ClientState){
			if($ClientState==CLIENT_VERIFIED){
				$this->SetDisabled(false);
			}
			elseif($OldState==CLIENT_VERIFIED){
				$this->SetDisabled(true);
			}
		}
		if($Ack)$this->DoSendStateResponse($ClientState,$Ack,$ClientIP, $ClientPort, $Id);		
	}
	protected function OnBeforeDestroy(){
		parent::OnBeforeDestroy();
// 		$this->SaveShareList([],false);
// 		if($MyParentID=intval($this->GetBuffer('LAST_CONNECTION_ID'))){
// 			if($this->GetStatus()!=201){
// 				if($s=@IPS_GetInstance($MyParentID)['InstanceStatus']){
// 					if($s==102){
// 						$this->DoSendDestroy();
// 					}
// 				}
// 			}
// 		}
		
		
	}
	protected function ApplyCleanUp(){
		$removeProfiles=[];
		foreach($this->LoadShareIndex() as $share){
			if($share[SHARE_LOCALID]>0 && $share[SHARE_TYPE]==OT_VARIABLE){
				if($v=@IPS_GetVariable($share[SHARE_LOCALID])){
					if(substr($v['VariableCustomProfile'],0,7)=='REMOTE_'){
						$removeProfiles[]=$v['VariableCustomProfile'];	
					}
				}
			}
		}
		parent::ApplyCleanUp();
		foreach($removeProfiles as $name){
			if(IPS_VariableProfileExists($name))IPS_DeleteVariableProfile($name);
		}
		
		
	}
	
	Protected function ApplyData(){
		parent::ApplyData();
		$shares=$this->LoadShareList();
		if(count($shares)==0)return $this->ApplyCleanUp();

		// Now detect changes from ShareIndex
 		$shareIndex = $this->LoadShareIndex();
 		$indexDiff=$this->GetSharesIndexDiff($shares,$shareIndex);
 		if($this->HasDiffChanges($indexDiff)){
 			
 			$this->ApplySharesIndexDiff($indexDiff,$shareIndex);
 			// check if linked ids changed
 			foreach($indexDiff[DIFF_UPDATE] as $update){
 				if(isset($update[SHARE_STATE]) && ($update[SHARE_STATE]==STATE_CONNECTED||$update[SHARE_STATE]==STATE_READY)){
   					$linkedIds=array_keys(array_filter($shareIndex,function($s){return $s[SHARE_LOCALID]>0;}));
 					$this->DoSendLinkedObjectIDsResponse($linkedIds);
 					break;
 				}
 			}

 		}
		
		
	}

	private function AfterDelete(){
		$fn=__DIR__.'/cleanup'.$this->InstanceID.'.data';
		if(file_exists($fn)){
			list($ident,$cat,$con)=explode(',', file_get_contents($fn));
			unlink($fn);
			if($con && ($s=@IPS_GetInstance($con)) && $s['InstanceStatus']==102){
				$buffer=$this->Encode_Buffer([DTYPE_DESTROY, 0,false,1,$ident],true).BUFFER_SUFFIX;
				CSCK_SendText($con,$buffer);	
			}
			if($cat && ($cat=@IPS_GetObject($cat)) && $cat['ObjectType']=OT_CATEGORY && $cat['HasChildren']){
				foreach($cat['ChildrenIDs'] as $deleteID){
					$this->DeleteObject($deleteID,true);
				}
			}
		}
	}
	private function SaveCleanupData($Clear=false){
		$fn=__DIR__.'/cleanup'.$this->InstanceID.'.data';
		if($Clear){
			if(file_exists($fn))unlink($fn);
			return;
		}
		$cat=$this->ReadPropertyInteger('CreateCategoryID');
		$con=intval($this->GetBuffer('LAST_CONNECTION_ID'));
		$ident=$this->ReadPropertyString('Identifier');
		file_put_contents($fn, "$ident,$cat,$con");
	}
	
	private function GetShareChangesRequest(array $MyShares=null){
		if(is_null($MyShares))$MyShares=$this->LoadShareIndex();
		return array_map(function($s){return [SHARE_DIR=>$s[SHARE_DIR]];},$MyShares);
	}
	
	private function SendConnectRequest(){
		$this->DoSendConnectRequest($this->GetShareChangesRequest());
	}
	
	
	private function ConnectAll(){
		$filter=array_filter($shares=$this->LoadShareList(),function($s){return $s[SHARE_LOCALID]==0;});
		$request=[];
		foreach($filter as $share){
			$this->CreateShare($share[SHARE_OBJ_ID], $request, $shares);
		}
		if(count($request)>0){
			$this->SaveShareList($shares);
			$this->DoSendObjectRequest($request,VOPT_DISABLED+VOPT_ICON+VOPT_PROFILE+VOPT_POSITION+VOPT_VALUE);
		}
		
	}
	private function DisconnectAll(){
		$filter=array_filter($shares=$this->LoadShareList(),function($s){return $s[SHARE_LOCALID]>0;});
		foreach($filter as $share){
			$this->DeleteShare($share[SHARE_OBJ_ID], $shares);
		}
		$this->SaveShareList($shares);
// 		$this->SendLinkedIDs();
		
	}
	private function SetDisabled($Disabled){
		$filter=array_filter($this->LoadShareList(),function($s){return $s[SHARE_LOCALID]>0;});	
		$buffer=$this->ReadPropertyString('LastDisableStates');
		if($Disabled && $buffer>'' || !$Disabled && $buffer=='')return;
		$lastDisabled=$Disabled?[]:json_decode($buffer,true);
		foreach($filter as $share){
			$localeID=$share[SHARE_LOCALID];
			$o=IPS_GetObject($localeID);
			if($Disabled){
				$lastDisabled[$localeID]=$o['ObjectIsDisabled'];
				$this->UnRegisterObjectMessages($localeID,$share[SHARE_TYPE]);
				IPS_SetDisabled($localeID, true);
			}elseif(isset($lastDisabled[$localeID])){
				IPS_SetDisabled($localeID, $lastDisabled[$localeID]);
			}
		}
		if(!$Disabled){
			IPS_Sleep(200);
			$linked=[];
			foreach($filter as $share){
				$localeID=$share[SHARE_LOCALID];
				if(isset($lastDisabled[$localeID])){
					$this->RegisterObjectMessages($localeID,$share[SHARE_TYPE]);
					$linked[]=$localeID;
				}
			}
			if(count($linked)!=count($lastDisabled)){ // Objects removed while client is disabled
// 				$this->SendLinkedIDs();
			}
		}
		$this->BlockApply();
		IPS_SetProperty($this->InstanceID, 'LastDisableStates', $Disabled?json_encode($lastDisabled):'');
		IPS_ApplyChanges($this->InstanceID);
		$this->UnBlockApply();
	}
// 	private function RemoveUnusedProfile(int $LocaleID, int $RemoteID){
// 		#TODO Remove unused Profiles
// 		if($v=@IPS_GetVariable($LocaleID)){
			
// 		}
// 	}
	
	private function DeleteObject(int $LocaleID, $Recrusive=false){
		$ok=true;
		if($o=@IPS_GetObject($LocaleID)){
			switch($o['ObjectType']){
				case OT_CATEGORY: $ok=false;//IPS_DeleteCategory($LocaleID);break; 
				case OT_INSTANCE: 
					if($o['HasChildren']){
						if($Recrusive){
							foreach($o['ChildrenIDs'] as $id){
								if(!($ok=$this->DeleteObject($id,$Recrusive)))break;
							}
						}
						else $ok=false;
					}
					if($ok)$ok=IPS_DeleteInstance($LocaleID);
					break;
				case OT_VARIABLE:$ok=IPS_DeleteVariable($LocaleID);break;
				case OT_SCRIPT	: $ok=IPS_DeleteScript($LocaleID);break;
				case OT_EVENT	: $ok=IPS_DeleteEvent($LocaleID);break;
				case OT_MEDIA	: $ok=IPS_DeleteMedia($LocaleID);break;
				case OT_LINK	: $ok=IPS_DeleteLink($LocaleID);break;
			}
		}
		return $ok;
	}
	private function CreateShare($ObjectID , array &$Requests, array &$ShareList = null){
		if(is_null($ShareList))$shares=$this->LoadShareList();else $shares=&$ShareList;
		$root=array_filter($shares,function($s)use($ObjectID){return $s[SHARE_OBJ_ID]==$ObjectID;});
		$rid=key($root);
		$created=false;
		if(is_null($rid) ||$root[$rid][SHARE_LOCALID]!=0 ){
			return false;
		}
		$share=$root[$rid];
		$shareIndex=$this->LoadShareIndex();
		
		if($share[SHARE_TYPE]==OT_INSTANCE){
			$created=true;			
			$shares[$rid][SHARE_LOCALID]=$rootID=IPS_CreateInstance('{485D0419-BE97-4548-AA9C-C083EB82E61E}'); //Dummy Instance
 			$shares[$rid][SHARE_STATE]=STATE_CREATE;
			$shareIndex[$ObjectID]=$shares[$rid];
			if(is_array($Requests)&& !in_array($ObjectID,$Requests))$Requests[]=$ObjectID;
		 	IPS_SetName($rootID,$share[SHARE_NAME]);
			IPS_SetParent($rootID, $this->ReadPropertyInteger('CreateCategoryID'));
			$childs=array_filter($shares,function($s)use($ObjectID){return $s[SHARE_PARENT]==$ObjectID;});
			foreach($childs as $cid=>$child){
				if($child[SHARE_TYPE]==OT_VARIABLE){
					$shares[$cid][SHARE_LOCALID]=$childID=IPS_CreateVariable($child[SHARE_VTYPE]);
					$shares[$cid][SHARE_STATE]=STATE_CREATE;
 					$shareIndex[$child[SHARE_OBJ_ID]]=$shares[$cid];
 					IPS_SetParent($childID, $rootID);
					IPS_SetName($childID,$child[SHARE_NAME]);
					if(is_array($Requests)&& !in_array($child[SHARE_OBJ_ID],$Requests))$Requests[]=$child[SHARE_OBJ_ID];
				}
			}
		}
		if($share[SHARE_TYPE]==OT_VARIABLE){
			$created=true;
			$shares[$rid][SHARE_LOCALID]=$rootID=IPS_CreateVariable($share[SHARE_VTYPE]);
			$shares[$rid][SHARE_STATE]=STATE_CREATE;
			$shareIndex[$ObjectID]=$shares[$rid];
			IPS_SetName($rootID,$share[SHARE_NAME]);
			IPS_SetParent($rootID, $this->ReadPropertyInteger('CreateCategoryID'));
			if(is_array($Requests)&& !in_array($ObjectID,$Requests))$Requests[]=$ObjectID;
		}
		if($created)$this->SaveShareIndex($shareIndex,false);
		if(is_null($ShareList))$this->SaveShareList($shares);
		return $created;
	}
	private function DeleteShare($ObjectID, array &$ShareList = null){
		if(is_null($ShareList))$shares=$this->LoadShareList();else $shares=&$ShareList;
		$root=array_filter($shares,function($s)use($ObjectID){return $s[SHARE_OBJ_ID]==$ObjectID;});
		$rid=key($root);
		$deleted=false;
		if(is_null($rid) ||$root[$rid][SHARE_LOCALID]==0 ){
			return false;
		}
		$share=$root[$rid];
		$shareIndex=$this->LoadShareIndex();
		if($share[SHARE_TYPE]==OT_INSTANCE){
			$deleted=true;
			$childs=array_filter($shares,function($s)use($ObjectID){return $s[SHARE_PARENT]==$ObjectID;});
			foreach($childs as $cid=>$child){
				if(empty($child[SHARE_LOCALID]))continue;
				if($child[SHARE_TYPE]==OT_VARIABLE){
					$this->OnShareDelete($child);
					$shareIndex[$child[SHARE_OBJ_ID]][SHARE_LOCALID]=$shares[$cid][SHARE_LOCALID]=0;
					$shares[$cid][SHARE_STATE]=STATE_READY;
				}
			}
			$this->OnShareDelete($share);
			$shareIndex[$ObjectID][SHARE_LOCALID]=$shares[$rid][SHARE_LOCALID]=0;
			$shares[$rid][SHARE_STATE]=STATE_READY;
		}
		if($share[SHARE_TYPE]==OT_VARIABLE){
			$deleted=true;
			$this->OnShareDelete($share);
			$shareIndex[$ObjectID][SHARE_LOCALID]=$shares[$rid][SHARE_LOCALID]=0;
			$shares[$rid][SHARE_STATE]=STATE_READY;
		}
		
		if($deleted)$this->SaveShareIndex($shareIndex,false);
		if(is_null($ShareList))$this->SaveShareList($shares);
		return $deleted;
	}
	private function UpdateVariableActionScript($LocaleID, $Direction, $Enabled){
		if(IPS_VariableExists($LocaleID) && ($v=@IPS_GetVariable($LocaleID))){
			if(empty($v['VariableCustomProfile']))return;
			$action=$v['VariableCustomAction'];
			$syncValid=$this->isValidSyncMode($Direction);
			if(!$Enabled && $action && !$syncValid){
				IPS_SetVariableCustomAction($LocaleID, 0);
				/*DBG*/if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->SendDebug(__FUNCTION__.':'.__LINE__,"Disable action for $LocaleID",0);
			}
			elseif($Enabled && !$action && $syncValid){
				IPS_SetVariableCustomAction($LocaleID, $this->getActionScriptID());
				/*DBG*/if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->SendDebug(__FUNCTION__.':'.__LINE__,"Enable action for $LocaleID",0);
			}			
		}
	}
	private function GetActionScriptID(){
		if(!($scriptID=@$this->GetIDForIdent('VariableAction'))){
			$scriptID=$this->RegisterScript("VariableAction", "Custom Action Script","<?php SetValue(\$_IPS['VARIABLE'],\$_IPS['VALUE']);?>", 0);
			IPS_SetHidden($scriptID, true);
		}
		return $scriptID;
	}
	
}
?>