<?php
#TODO Update ClientStates when Client deleted in ClientList

require_once __DIR__.'/../libs/RemoteIOModule.inc';
class BaseModuleClients extends RemoteIOModule {
	public function Create() {
		parent::Create();
		$this->RegisterPropertyBoolean('ShowClientList', true);
		$this->RegisterPropertyBoolean('ShowClientStates', true);
		$this->RegisterPropertyBoolean('AllowNewClients', true);
		$this->RegisterPropertyString('ClientList', '');
 		if(KERNEL_VERSION>5)$this->RegisterAttributeString('ClientStates', '');
		else $this->SetBuffer('ClientStates', '');
		$this->SetBuffer('ClientsAllowed','');
	}
	public function RequestAction($Ident, $Value){
			if($Ident=='CHECK_CLIENTS'){
 				$this->CheckClientsAlive(true);
			}
			else parent::RequestAction($Ident, $Value);
	
	}
	// -----------------------------------------
	// -  Form Maintain
	// -----------------------------------------
	protected function GetFormConfigItems(int $FormVersion){
		$items=parent::GetFormConfigItems($FormVersion);
		$items[]=["type"=>"CheckBox", "name"=>"ShowClientList","caption"=>"Show Client list"];
		$items[]=["type"=>"CheckBox", "name"=>"ShowClientStates","caption"=>"Show Client states"];
		$items[]=["type"=>"CheckBox", "name"=>"AllowNewClients","caption"=>"Allow connect for new Clients"];
		return $items;
	}
	protected function GetFormList(int $FormVersion){
		$items=parent::GetFormList($FormVersion);
		if($this->ReadPropertyBoolean('ShowClientList')){
			$caption= $FormVersion<5?'label':'caption';
			$width	= $FormVersion<5?'100px':'auto';
			$values=$this->LoadClientList();
			foreach($values as &$v){
	 			$v['Linked']=json_encode($v['LinkedIDs']);
			}
			$items[]=["type"=>"List",	  "name"=>"ClientList", "caption" => "Clients","add"=> false,"delete"=> true,"rowCount"=>4,
						"columns"=>[
							[ "name"=>'Ip', 		$caption=> "IP",		"width"=>"120px","save"=>true], 
							[ "name"=>'Name', 		$caption=> "Name",		"width"=>"120px", "edit"=> ["type"=> "ValidationTextBox","caption"=>"Name"]], 
							[ "name"=>'Allowed',	$caption=> "Enabled",	"width"=>"80px", "edit"=> ["type"=> "CheckBox","caption"=>"Client Allowed"]], 
							[ "name"=>'Identifier',	$caption=> "Identifier","width"=>"90px", "save"=>true], 
							[ "name"=>'Linked',	 	$caption=> "Linkeds IDs","width"=>$width,"save"=>false], 	
							[ "name"=>'LinkedIDs', 	$caption=> "",			"width"=>"0px", "visible"=>false, "save"=>true], 
						],
						"values"=>$values
			];
		}
		return $items;
	}
	// -----------------------------------------
	// -  Override Events Maintain
	// -----------------------------------------
	protected function OnKernelReady(){
		$this->CreateClientsAllowed();
		parent::OnKernelReady();
	}
	protected function OnControlTimer($Mode){
		parent::OnControlTimer($Mode);
		$this->CheckClientsAlive();
	}
	protected function OnPongReceived($PongData, $Ack, $Id, $ClientIP, $ClientPort){
		parent::OnPongReceived($PongData, $Ack, $Id, $ClientIP, $ClientPort);
		$this->UpdateClientState($ClientIP, $ClientPort,true,$PongData);
	}
	protected function OnPingReceived($PingData, $Ack, $Id, $ClientIP, $ClientPort){
		parent::OnPingReceived($PingData, $Ack, $Id, $ClientIP, $ClientPort);
		$clientState=$this->UpdateClientState($ClientIP, $ClientPort,true,$PingData);
		if($Ack)$this->DoSendPong($clientState,false,$ClientIP, $ClientPort,$Id);
	}
	protected function OnLinkedObjectIDsResponse($LinkedObjectIDs, $Ack, $Id, $ClientIP, $ClientPort){
		parent::OnLinkedObjectIDsResponse($LinkedObjectIDs, $Ack, $Id, $ClientIP, $ClientPort);
		$clientState=$this->UpdateClientState($ClientIP, $ClientPort,true,null,$LinkedObjectIDs);
		if($clientState==CLIENT_NOT_ALOWED){
			$this->DoSendStateResponse($clientState);			
		}
	}
	protected function OnStateResponse($Ident, $ClientState, $Ack, $ClientIP, $ClientPort, $Id){
		parent::OnStateResponse($Ident, $ClientState, $Ack, $ClientIP, $ClientPort, $Id);
		$state=$this->UpdateClientState($ClientIP, $ClientPort,true,$Ident);
		if($state!=$ClientState){
			$this->UpdateShareStatesByClients();
		}
	}
	// -----------------------------------------
	// -  Override Data Sending to find ClientIPs
	// -----------------------------------------
	protected function DoSendMessage(array $Message, $Ack=false,$ClientIP='',$ClientPort=0,$Id=0){
		if(empty($ClientIP)){
			foreach($this->GetClientsToSend([$Message[0]]) as $ClientIP=>$ClientPort){
				parent::DoSendMessage($Message,$Ack,$ClientIP,$ClientPort,$Id);
			}
		}else parent::DoSendMessage($Message,$Ack,$ClientIP,$ClientPort,$Id);
	}
	protected function DoSendChangesResponse($ShareChangesResponse,$Ack=false,$ClientIP='',$ClientPort=0,$Id=0){
		if(empty($ClientIP)){
			foreach($this->GetClientsToSend(
					$this->HasDiffUpdatesOnly($ShareChangesResponse)?
					array_keys($ShareChangesResponse[DIFF_DELETE]):[]) as $ClientIP=>$ClientPort){
				parent::DoSendChangesResponse($ShareChangesResponse,$Ack,$ClientIP,$ClientPort,$Id);			
			}
		} else parent::DoSendChangesResponse($ShareChangesResponse,$Ack,$ClientIP,$ClientPort,$Id);
	}
	
	// -----------------------------------------
	// -  Client Events Maintain
	// -----------------------------------------
	protected function OnClientSocketConnected($ClientIP, $ClientPort){
		parent::OnClientSocketConnected($ClientIP, $ClientPort);
		$this->UpdateClientState($ClientIP, $ClientPort,true);
	}
	protected function OnClientSocketDisConnected($ClientIP, $ClientPort){
		parent::OnClientSocketDisConnected($ClientIP, $ClientPort);
		$this->UpdateClientState($ClientIP, $ClientPort,false);
	}
	
	protected function OnClientIPChanged(string $OldIP,string $NewIP){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("IP: $OldIP to New IP: $NewIP",__FUNCTION__);
		$allowed=$this->LoadClientsAllowed();
		if(isset($allowed[$OldIP])){
			$allowed[$NewIP]=$allowed[$OldIP];
			unset($allowed[$OldIP]);
			$this->SaveClientsAllowed($allowed);
		}
		
	}
	protected function OnClientIdentifierChanged(string $ClientIP, string $OldIdentifier,string $NewIdentifier){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("IP: $ClientIP  from $OldIdentifier to $NewIdentifier",__FUNCTION__);
	}
	
	protected function OnClientVerification(string $ClientIP, string $Identifier){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("ClientIP: $ClientIP Identifier: $Identifier",__FUNCTION__);
	}
	protected function OnClientStateChanged(string $ClientIP, int $OldState, int $NewState){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$so=$this->TranslatedClientState($OldState);
			$sn=$this->TranslatedClientState($NewState);
			$this->Debug("ClientIP: $ClientIP from $so ($OldState) to $sn ($NewState)",__FUNCTION__);
		}
		if($OldState==CLIENT_VERIFIED || $NewState==CLIENT_VERIFIED){
			$this->UpdateShareStatesByClients();
		}
	}
	protected function OnClientLinkedIDsChanged(string $ClientIP, array $OldIDs, array $NewIDs){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$this->Debug("ClientIP: $ClientIP from ".json_encode($OldIDs). " to ".json_encode($NewIDs),__FUNCTION__);
		}
		$this->UpdateShareStatesByClients();
	}
	// -----------------------------------------
	// -  Clients Maintain
	// -----------------------------------------
	private $clientsAllowed=null;
	protected function CreateClientsAllowed(){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("execute",__FUNCTION__);
		$this->clientsAllowed=[];
		foreach($this->LoadClientList() as $client){
			$this->clientsAllowed[$client['Ip']]=$client['Allowed'];
		}
		$this->SetBuffer('ClientsAllowed',json_encode($this->clientsAllowed));
		return $this->clientsAllowed;
		
	}
	protected function LoadClientsAllowed(){
		if(!is_null($this->clientsAllowed)){
			if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Return from cache => ".json_encode($this->clientsAllowed),__FUNCTION__);				
			return $this->clientsAllowed;
		}
		$this->clientsAllowed=!is_null($l=json_decode($this->GetBuffer('ClientsAllowed'),true))?$l:$this->CreateClientsAllowed(); 
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug( json_encode($this->clientsAllowed), __FUNCTION__);				
		return $this->clientsAllowed;
	}
	protected function SaveClientsAllowed(array $ClientsAllowed){
		$encoded=json_encode($ClientsAllowed);
		if($encoded!=json_encode($this->LoadClientsAllowed())){
			$this->clientsAllowed=$ClientsAllowed;
			if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug($encoded,__FUNCTION__);		
			elseif(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Saved", __FUNCTION__);		
			$this->SetBuffer('ClientsAllowed',$encoded);
		}
	}
	protected function UpdateClientsAllowed(){
		$allowed=$this->LoadClientsAllowed();
		$changed=false;
		foreach($this->LoadClientList() as $client){
			$clientIP=$client['Ip'];
			if(!isset($allowed[$clientIP])){
				$allowed[$clientIP]=$client['Allowed'];
				$changed=true;
			}
			elseif($allowed[$clientIP]!=$client['Allowed']){
				$changed=true;
				$clientStates=$this->LoadClientStates();
				$clientPort=$clientStates[$clientIP]['Port'];
				$allowed[$clientIP]=$client['Allowed'];
				$state=!$client['Allowed']?CLIENT_NOT_ALOWED:(empty($client['Identifier'])?CLIENT_CONNECTED:CLIENT_VERIFIED);
				$this->DoSendStateResponse($state,true,$clientIP,$clientPort);				
			}
		}
		if($changed)$this->SaveClientsAllowed($allowed);
	}
	
	private $clientList=null;
	protected function LoadClientList(){
		if(!is_null($this->clientList)){
			if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Return from cache => ".json_encode($this->clientList),__FUNCTION__);				
			return $this->clientList;
		}
		 $this->clientList=($l=json_decode($this->ReadPropertyString('ClientList'),true))?$l:[]; 
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug( json_encode($this->clientList), __FUNCTION__);				
		return $this->clientList;
	}
	protected function SaveClientList(array $ClientList){
		$encoded=json_encode($ClientList);
		if($encoded!=json_encode($this->LoadClientList())){
			$this->clientList=$ClientList;
			if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug($encoded, __FUNCTION__);		
			else/*DBG*/if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug( "Saved", __FUNCTION__);		
 			$this->BlockApply();
			IPS_SetProperty($this->InstanceID, 'ClientList', $encoded);
			IPS_ApplyChanges($this->InstanceID);
 			$this->UnBlockApply();
		} 
	}
	protected function FindClientByIpOrIdentifier(string $ClientIP, string $Identifier=null, array $ClientList=null){
		if(is_null($ClientList))$ClientList=$this->LoadClientList();
		if($Identifier){
			$found=array_filter($ClientList,function($c)use($ClientIP, $Identifier){
				return ($c['Ip']==$ClientIP || (!empty($c['Identifier'])&& $c['Identifier']==$Identifier));
			});
		}
		else $found=array_filter($ClientList,function($c)use($ClientIP){return $c['Ip']==$ClientIP;});
		if(count($found)>1){
			uasort($found, function($c1,$c2){
				if($c1['Identifier'] && $c2['Identifier'])return 0;
				return ($c1['Identifier'])?-1:1;
			});
		}
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Returns => ".json_encode($found),__FUNCTION__);				
		
		return empty($found)?null:$found;
	}
	protected function GetClientsToSend(array $ObjectIDs){
		$clientList		= $this->LoadClientList();
		$clientStates	= $this->LoadClientStates();
		$list=[];
		foreach($clientList as $client){
			if($client['Allowed'] &&  empty($list[$client['Ip']]) && (empty($ObjectIDs) || !empty(array_intersect ( $ObjectIDs , $client['LinkedIDs'])))){
				$state=$clientStates[$client['Ip']];
				
				$list[$client['Ip']]=$state['Port'];
			}
		}
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Returns => ".json_encode($list),__FUNCTION__);
		return $list;
	}
	
	private $clientStates=null;
	protected function LoadClientStates(){
		if(!is_null($this->clientStates)){
			if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Return from cache => ".json_encode($this->clientStates),__FUNCTION__);				
			return $this->clientStates;
		}
		 $this->clientStates=($l=json_decode(
				KERNEL_VERSION > 5.0 ? $this->ReadAttributeString('ClientStates'):
				$this->GetBuffer('ClientStates'),true))?$l:[]; 
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug( json_encode($this->clientStates), __FUNCTION__);				
		return $this->clientStates;
	}
	protected function SaveClientStates(array $ClientStates){
		$encoded=json_encode($ClientStates);
		if($encoded!=json_encode($this->LoadClientStates())){
			$this->clientStates=$ClientStates;
			if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug($encoded,__FUNCTION__);		
			elseif(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Saved", __FUNCTION__);		
			if(KERNEL_VERSION > 5.0)
				$this->WriteAttributeString('ClientStates',$encoded);
			else
				$this->SetBuffer('ClientStates',$encoded);
		}
	}
	Protected function UpdateClientState(string $ClientIP, int $ClientPort, bool $Connected=true,string $Identifier=null, array $ConnectedIDs=null){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$this->Debug("Ident: $Identifier ClientIP: $ClientIP ClientPort: $ClientPort Connected: ".$this->Bool2Str($Connected)." IDs => ".json_encode($ConnectedIDs),__FUNCTION__);
		}
		$clientStates=$this->LoadClientStates();
		$clientListChanged=$oldIP=$ipChanged=$linkedIDsChanged=false;
		$OldClientState=empty($clientStates[$ClientIP])?CLIENT_DISCONNECTED:$clientStates[$ClientIP]['State'];
		if(empty($clientStates[$ClientIP]) || !empty($Identifier) || is_array($ConnectedIDs)){
			if(!empty($Identifier) || is_array($ConnectedIDs)){
				$clientList=$this->LoadClientList();
				if($found=$this->FindClientByIpOrIdentifier($ClientIP,$Identifier,$clientList)){
					if(count($found)>1){
						foreach(array_keys($found) as $index){
							if($clientListChanged){
								unset($clientList[$index]);
								unset($found[$index]);
							}
							$clientListChanged=true;
						}
					}
					$index=key($found);
					if(!empty($Identifier) && $clientList[$index]['Identifier'] !=$Identifier){
						$this->OnClientIdentifierChanged($ClientIP, $clientList[$index]['Identifier'], $Identifier);
						$clientList[$index]['Identifier']=$Identifier;
						if(!is_array($ConnectedIDs))$ConnectedIDs=[];
						$clientListChanged=true;
					}
					
					if($ipChanged=($oldIP=$clientList[$index]['Ip'])!=$ClientIP){
						if(isset($clientStates[$oldIP])){
							$clientStates[$ClientIP]=$clientStates[$oldIP];
							unset($clientStates[$oldIP]);
						}
						$clientList[$index]['Ip']=$ClientIP;
						$clientListChanged=true;
					}
					if(is_array($ConnectedIDs) &&  $clientList[$index]['LinkedIDs']!=$ConnectedIDs){
						$oldLinkedIDs=$clientList[$index]['LinkedIDs'];
						$clientList[$index]['LinkedIDs']=$ConnectedIDs;
						$clientListChanged=$linkedIDsChanged=true;
					}
					$clientStates[$ClientIP]['State']=!$Connected?CLIENT_DISCONNECTED:($clientList[$index]['Allowed']?CLIENT_VERIFIED:CLIENT_NOT_ALOWED);
					if($clientListChanged)$clientList=array_values($clientList);
				} else {
					$clientList[]=[
						'Ip'=>$ClientIP,
						'Name'=>$Identifier,
						'Allowed'=>$a=($this->ReadPropertyBoolean('AllowNewClients')||$this->OnClientVerification($ClientIP, $Identifier)!==false),
						'Identifier'=>$Identifier,
						'LinkedIDs'=>is_array($ConnectedIDs)?$ConnectedIDs:[]	
					];
					$allowed=$this->LoadClientsAllowed();
					$allowed[$ClientIP]=$a;
					$this->SaveClientsAllowed($allowed);
					$clientListChanged=true;
					$clientStates[$ClientIP]['State']=!$Connected?CLIENT_DISCONNECTED:($a?CLIENT_CONNECTED:CLIENT_NOT_ALOWED);
				}
			}else {
				$clientStates[$ClientIP]=[
					'State'=>$Connected?CLIENT_CONNECTED:CLIENT_DISCONNECTED,
				];				
			}
		} elseif($Connected){
			$clientStates[$ClientIP]['State']=isset($clientStates[$ClientIP]['save_state'])?$clientStates[$ClientIP]['save_state']:CLIENT_CONNECTED;
			unset($clientStates[$ClientIP]['save_state']);
		} elseif(!isset($clientStates[$ClientIP]['save_state'])){
			$clientStates[$ClientIP]['save_state']=$clientStates[$ClientIP]['State'];
			$clientStates[$ClientIP]['State']=CLIENT_DISCONNECTED;
		}		
		$clientStates[$ClientIP]['Port']=$ClientPort;
		$clientStates[$ClientIP]['Time']=time();
		$this->SaveClientStates($clientStates);
		if($clientListChanged)$this->SaveClientList($clientList);
		if($ipChanged)$this->OnClientIPChanged($oldIP,$ClientIP);
		if($OldClientState!=$clientStates[$ClientIP]['State']){
			$this->OnClientStateChanged($ClientIP, $OldClientState, $clientStates[$ClientIP]['State']);
		}
		if($linkedIDsChanged){
			$this->OnClientLinkedIDsChanged($ClientIP,$oldLinkedIDs,$ConnectedIDs);
		}
		return $clientStates[$ClientIP]['State'];
	}

	
	private function UpdateShareStatesByClients(){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$this->Debug("execute",__FUNCTION__);
		}
		$clientList = $this->LoadClientList();
		$shareList	= $this->LoadShareList();
		$clientStates = $this->LoadClientStates();
		$regIDs		= [];
		foreach($clientList as $client){
			$ip=$client['Ip'];
			if($client['Allowed']){
				if(empty($clientStates[$ip])|| $clientStates[$ip]['State']==CLIENT_VERIFIED)
					$regIDs=array_merge($regIDs,$client['LinkedIDs']);
			}
		}
		$regIDs=array_keys(array_flip($regIDs));
		$changed=false;
		foreach($shareList as &$share){
			if($share[SHARE_STATE]==STATE_CONNECTED && !in_array($share[SHARE_OBJ_ID],$regIDs )){
				$share[SHARE_STATE]=STATE_READY;$changed=true;
			}elseif($share[SHARE_STATE]!=STATE_CONNECTED && in_array($share[SHARE_OBJ_ID],$regIDs) ){
				$share[SHARE_STATE]=STATE_CONNECTED;$changed=true;
			}
		}
		if($changed)$this->SaveShareList($shareList);
	}
	
	private $ValidClientAliveTime = 3600; // 1 Stunde
	private function CheckClientsAlive($Force=false){
		$now 			= time();
		$clientStates	= $this->LoadClientStates();
		$filter			= array_filter($clientStates,function($c)use($now,$Force){return $Force||$c['Time']+$this->ValidClientAliveTime < $now;});
		if(($missing=count($filter))>0){
			$changes=[];
			if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("( $missing ) clients missing",__FUNCTION__);
			foreach($filter as $clientIP=>$item){
				$state=$item['State'];
				if($state==CLIENT_VERIFIED){
					$state=CLIENT_CONNECTED;
				}
				elseif($state==CLIENT_CONNECTED){
					$state=CLIENT_DISCONNECTED;
				}
				elseif($state==CLIENT_DISCONNECTED){
					$state=CLIENT_FAILS;
				}
				elseif($state==CLIENT_FAILS){
					$state=CLIENT_CLEANUP;
				}
				elseif($state==CLIENT_CLEANUP){
// 					$states[$clientIP]['State']=CLIENT_CONNECTED;
				}
				if($clientStates[$clientIP]['State']!=$state){
					if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("( $clientIP ) missing. Set State to => ".$this->TranslatedClientState($state),__FUNCTION__);
					if($state!=CLIENT_CLEANUP)$this->DoSendPing($state,true,$clientIP,$clientStates[$clientIP]['Port']);
					$changes[$clientIP]=$clientStates[$clientIP]['State'];
					$clientStates[$clientIP]['State']=$state;
				}
			}
			if(count($changes)>0){
				$this->SaveClientStates($clientStates);
				foreach($changes as $clientIP=>$OldState){
					$this->OnClientStateChanged($clientIP, $OldState, $clientStates[$clientIP]['State']);
				}
			}
		}
		else if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("All ( ".count($clientStates)." ) clients uptodate",__FUNCTION__);
	}
}	
class BaseModuleServer extends BaseModuleClients {

		
	// -----------------------------------------
	// -  Form Maintain
	// -----------------------------------------
	protected function GetFormConfigItems(int $FormVersion){
		$items=parent::GetFormConfigItems($FormVersion);
		$items[]=["type"=>"CheckBox", "name"=>"AutoSendShares","caption"=>"Auto Send Shares if changed"];
		return $items;
	}
	protected function GetFormButtons(int $FormVersion){
		$items=parent::GetFormButtons($FormVersion);
		$caption=$FormVersion<5?'label':'caption';
		$msgSend=$this->Translate("Shares have been sent");
		$msgCheck=$this->Translate("Shares were checked");
		$msgClnt=$this->Translate("Clients were checked");
		$items[]=["type"=>"Button",$caption=>"Send Shares","onClick"=>"if(IPS_RequestAction($this->InstanceID,'SEND_SHARES',0))echo '$msgSend';"];
		$items[]=["type"=>"Button",$caption=>"Check Shares","onClick"=>"if(IPS_RequestAction($this->InstanceID,'CHECK_SHARES',0))echo '$msgCheck';"];
		$items[]=["type"=>"Button",$caption=>"Check Clients","onClick"=>"if(IPS_RequestAction($this->InstanceID,'CHECK_CLIENTS',0))echo '$msgClnt';"];
		return $items;
	}
	protected function GetFormList(int $FormVersion){
		$values		= $this->GetFormListValues($FormVersion);
		$caption	= $FormVersion<5?'label':'caption';
		$width		= $FormVersion<5?'100px':'auto';
		$nextID 	= count($values)+1;
		$items		= [];
		if($FormVersion!=5){
			$id=	[ "name"=> 'id', $caption=> "ID", "width"=> "30px", "add"=>0, "save"=>true];
			$parent=[ "name"=> 'parent', $caption=> "Parent", "width"=> "30px","add"=>'',"save"=>true];
		}else {
			$id=	["name"=>"id",	  $caption=> "", "width"=>"0px","add"=>$nextID,"save"=>true,"visible"=>false];
			$parent=["name"=>'parent', $caption=> "", "width"=>"0px","add"=>'',"save"=>true,"visible"=>false];
		}
		$items[]=[ "type"=> $FormVersion!=5?"List":"Tree","name"=> "ShareList", "caption"=> "Shares","add"=> true,"delete"=> true,
			    "columns"=> [ $id,$parent,
			    	[ "name"=> SHARE_OBJ_ID, $caption=> "Object ID","width"=> $width,"add"=> 0, "edit"=> ["type"=> "SelectObject","caption"=>"Variable or Module ID"]], 
			    	[ "name"=> "type", $caption=> "Object type", "width"=> "120px", "add"=> "", "save"=> false],
			    	[ "name"=> SHARE_DIR, $caption=> "Direction", "width"=> "150px", "add"=> 3, "edit"=> [
					        "type"=>"Select", "caption"=>"Select direction", "options"=> [
								["value"=>DIR_NONE, "label"=>self::DIR_NAMES[DIR_NONE]],
								["value"=>DIR_STOC, "label"=>self::DIR_NAMES[DIR_STOC]],
								["value"=>DIR_CTOS, "label"=>self::DIR_NAMES[DIR_CTOS]],
								["value"=>DIR_ALL, "label"=>self::DIR_NAMES[DIR_ALL]]
							]
			        	]
			    	], 
			    	[ "name"=> "state", $caption=> "State", "width"=> "100px", "add"=> $this->Translate("New"), "save"=> false],
			    	[ "name"=> SHARE_TYPE,  $caption=> "", "width"=> "0px", "add"=> 0, "save"=> true,"visible"=>false],
			    	[ "name"=> SHARE_STATE, $caption=> "", "width"=> "0px", "add"=> 0, "save"=> true,"visible"=>false],
			    	[ "name"=> SHARE_PARENT,$caption=> "", "width"=> "0px", "add"=> 0, "save"=> true,"visible"=>false],
			    	[ "name"=> SHARE_VTYPE, $caption=> "", "width"=> "0px", "add"=> 0, "save"=> true,"visible"=>false],
			    ],
				"values"=> $values
		];
	
		return array_merge($items,parent::GetFormList($FormVersion));
	}
	protected function GetFormCLientConfig(int $FormVersion){
		$clients	= $this->LoadClientList();
		$options	= array_map(function($c){return ['value'=>$c['Ip'],'label'=>$c['Name'].($c['Identifier']!=$c['Name']?' ('.$c['Identifier'].')':'')];},$clients); 
		$caption	= $FormVersion<5?'label':'caption';
		$confirm	= $this->Translate("Are you sure you want to configure the client?");
		$errSel		= $this->Translate("Please select a client first");
		$execute	= "if(is_null(\$CClient))echo '$errSel'; else IPS_RequestAction($this->InstanceID,'CONFIGURE_CLIENT',json_encode([
'Client'=>\$CClient, 'SendChangesOnly'=>\$CSendChangesOnly,
'DebugLevel'=>\$CDebugLevel, 'ControlInterval'=>\$CControlInterval,  'FormType'=>\$CFormType, 'EnableConfig'=>\$CEnableConfig,
'AllowRemoteCommands'=>\$CAllowRemoteCommands, 'EnableControlButtons'=>\$CEnableControlButtons, 'RemoveUnusedProfiles'=>\$CRemoveUnusedProfiles
]));";
		$button		= ["type"=>"Button", $caption=>"Configure Client", "onClick"=>$execute, "confirm"=>$confirm];
		$select		= ["type"=>"Select", "name"=>"CClient", "caption"=>"Select client", "options"=>$options];
		$optBool	= [["value"=>null, "label"=>""],["value"=>true, "label"=>"Yes"],["value"=>false, "label"=>"No"]];
		$items=[
				["type"=>"Select", "name"=>"CDebugLevel","caption"=>"Debug level", "options"=>[["value"=>'', "label"=>""],["value"=>DEBUG_NONE, "label"=>"None"],["value"=>DEBUG_ERROR, "label"=>"Errors"],["value"=>DEBUG_NOTICE, "label"=>"Notice"],["value"=>DEBUG_DETAIL, "label"=>"Detail"],["value"=>DEBUG_DEBUG, "label"=>"Debug"]]],
            	["type"=>"NumberSpinner", "name"=>"CControlInterval","caption"=>"Control Timer interval","suffix"=>"hours"],
				["type"=>"Select", "name"=>"CSendChangesOnly", "caption"=>"Send changes only", "options"=> $optBool],
  				["type"=>"Select", "name"=>"CAllowRemoteCommands", "caption"=>"Remote Commands", "options"=> [
						["value"=>null, "label"=>""],
						["value"=>COMMAND_ALLOW_NONE, "label"=>"Disabled"],
						["value"=>COMMAND_ALLOW_IPS, "label"=>"IPSymcon only"],
						["value"=>COMMAND_ALLOW_ALL, "label"=>"All"],
					]
				],
				["type"=>"Select", "name"=>"CEnableControlButtons", "caption"=>"Show control buttons", "options"=> $optBool],
				["type"=>"Select", "name"=>"CAutoCreateLinks", "caption"=>"Auto create links", "options"=> $optBool],
				["type"=>"Select", "name"=>"CRemoveUnusedProfiles", "caption"=>"Remove unused Profiles", "options"=> $optBool],
				["type"=>"Select", "name"=>"CEnableConfig", "caption"=>"Enable Configuration", "options"=> $optBool],
		];
		if($FormVersion>4){
			$items[]=["type"=>"Select", "name"=>"CFormType", "caption"=>"Form Mode", "options"=> [
					["value"=>'', "label"=>""],
					["value"=>5, "label"=>"Version 5"],
					["value"=>54, "label"=>"Version 4"],
					["value"=>FOMR_CONSOLE, "label"=>"Windows Console"],
				]
	       	];		
		}
		
		
		if($FormVersion<5){
			$items[]=$select;
			$items[]=$button;
		}else{
			$items[]=["type"=>"RowLayout", "items"=>[$select,$button]];
			$items=[["type"=>"PopupButton", "caption"=>"Configure Client", "popup"=> ["caption"=> "Configure Client", "items"=> $items]]];
		}
		return $items;
	}
	protected function GetFormActions(int $FormVersion){
		$items=parent::GetFormActions($FormVersion);
		if($this->ReadPropertyBoolean('ShowClientStates')){
	 		$states = $this->LoadClientStates();
			$caption=$FormVersion<5?'label':'caption';
			$width=$FormVersion<5?'200px':'auto';		
			
			foreach($states as $k=>&$v){
				$v['Ip']=$k;
				$v['Time']=Date('d.m - H:i:s',$v['Time']);
				$v['State']=self::CSTATE_NAMES[$v['State']];
	// 			$v['Allowed']=$this->TranslatedYesNo($v['Allowed']);
			}
			$items[]= [ "type"=> "List","name"=> "client", "caption"=> "Client States","add"=> false,"delete"=> false,"rowCount"=>4,
				    "columns"=> [
				    	[ "name"=> "Ip", $caption=> "IP", "width"=> $width],
				    	[ "name"=> "Port", $caption=> "Port", "width"=> "80px"],
				    	[ "name"=> "Time", $caption=> "Last connect", "width"=> "180px"], 
				    	[ "name"=> "State", $caption=> "State", "width"=> "140px"],
				    ],
					"values"=>array_values($states)
			];
		}
		if($cf=$this->GetFormCLientConfig($FormVersion)){
			$items = is_array($items)?array_merge($items,$cf):$cf;
		}
		return $items;
	}
}

class RemoteIOServer extends BaseModuleServer {
	protected $ConnectParentGuid='{8062CF2B-600E-41D6-AD4B-1BA66C32D6ED}';
	protected $IsServer = true;
	public function Create() {
		parent::Create();
		$this->RegisterPropertyBoolean('AutoSendShares', false);
	}

	public function RequestAction($Ident, $Value){
		if($Ident=='SEND_SHARES'){
				$this->SendShares();
			}
			elseif($Ident=='CHECK_SHARES'){
 // 				$this->CheckShareList();
			}
			elseif($Ident=='CREATE_REMOTE_OBJECT'){
				
	// 			$this->checkClientsAlive();
			}elseif($Ident=='CONFIGURE_CLIENT'){
				$this->ConfigureClient($Value);
	// 			$this->checkClientsAlive();
			}else parent::RequestAction($Ident, $Value);
	
	}
	

	// -----------------------------------------
	// -  Override Events Maintain
	// -----------------------------------------
	protected function OnBeforeObjectValueSet($LocaleID, $ValueOption, $NewValue){
		parent::OnBeforeObjectValueSet($LocaleID, $ValueOption, $NewValue);
		if($ValueOption != VOPT_VALUE ) return;
		if($LocaleID==0 || !($v=@IPS_GetVariable($LocaleID)))return [false,ERR_ObjectIsNoVariable,null];
		if($action=$v['VariableCustomAction']){
			if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug('RunScript => '.$action,__FUNCTION__);
			if(!IPS_ScriptExists($action)) return [false, ERR_ActionScriptNotExist,null];
			return [IPS_RunScriptEx($action, ['VARIABLE'=>$LocaleID,'VALUE'=>$NewValue]),ERR_ExecuteActionScript,null] ;
		}
		elseif($action=$v['VariableAction']){
			$checkModuleStatusRecrusive=null;
			$checkModuleStatusRecrusive=function($InstanceID)use(&$checkModuleStatusRecrusive){
				if(!($i=IPS_GetInstance($InstanceID)))return false;
				$status=$i['InstanceStatus'];
				if(REMOTEIO_DEBUG>DEBUG_NOTICE){
					$name=IPS_GetName($InstanceID);
					$this->Debug("($InstanceID) $name => $status",'SetLocalValue:CheckStatus');
				}
				if($status==102 && $i['ConnectionID']>0){
					$status=$checkModuleStatusRecrusive($i['ConnectionID']);
				}
				return $status;
			};
			$o=IPS_GetObject($LocaleID);
			$name=$o['ObjectName'];
			$i=$checkModuleStatusRecrusive($o['ParentID']);
			if($i===false){
				 if(REMOTEIO_DEBUG>0)$this->Debug("Parent \"({$o['ParentID']}) $name\" is not a Module",__FUNCTION__);
				return [false,ERR_ObjectIsNoInstance,null];
			}
			elseif($i!=102){
				if(REMOTEIO_DEBUG>0)$this->Debug("Instance status $i from \"({$o['ParentID']}) $name\" or parents, is invalid",__FUNCTION__);
				return [false,ERR_InvalidInstanceStatus,$i];
			}
			else {
				if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("RequestAction for \"({$o['ParentID']}) $name\" Ident: {$o['ObjectIdent']}  => ".$NewValue,__FUNCTION__);
				IPS_RequestAction($o['ParentID'], $o['ObjectIdent'], $NewValue);
				return [true,ERR_ExecuteInstanceAction,null];
			}
		}
		else{
			if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->SendDebug("SetValue for $LocaleID to =>".$NewValue,__FUNCTION__);
			return [(bool)@SetValue($LocaleID,$NewValue),ERR_ExecuteSetValue,null];
		}
		
	}
	protected function OnConnectRequest(string $Ident, array $ShareIndex,string $ClientIP,int $ClientPort ,int $Id){
		parent::OnConnectRequest($Ident, $ShareIndex, $ClientIP, $ClientPort, $Id);
		$clientState=$this->UpdateClientState($ClientIP, $ClientPort,true,$Ident);
		$diff=$clientState==CLIENT_VERIFIED?$this->GetShareChangesResponse($ShareIndex):[];
		$this->DoSendConnectResponse($clientState, $diff,false,$ClientIP,$ClientPort,$Id);
	}
	
	protected function OnShareAdd(&$Share){
		parent::OnShareAdd($Share);
		$this->RegisterObject($Share[SHARE_OBJ_ID],$Share[SHARE_TYPE]);
		if(KERNEL_VERSION>5)$this->RegisterReference($Share[SHARE_OBJ_ID]);
		if($Share[SHARE_STATE]==STATE_CONNECTED)$this->OnShareEnabled($Share);
	}
	protected function OnShareDelete($Share){
		parent::OnShareDelete($Share);
		if(KERNEL_VERSION>5)$this->UnRegisterReference($Share[SHARE_OBJ_ID]);
		$this->UnRegisterObject($Share[SHARE_OBJ_ID],$Share[SHARE_TYPE]);
		if($Share[SHARE_STATE]==STATE_CONNECTED)$this->OnShareDisabled($Share);
	}
	protected function OnShareEnabled(array $Share){
		parent::OnShareEnabled($Share);
		if($this->isValidSyncMode($Share[SHARE_DIR]))
			$this->RegisterObjectMessages($Share[SHARE_OBJ_ID],$Share[SHARE_TYPE]);
		else $this->UnRegisterObjectMessages($Share[SHARE_OBJ_ID],$Share[SHARE_TYPE]);
	}
	protected function OnShareDisabled(array $Share){
		parent::OnShareDisabled($Share);
		$this->UnRegisterObjectMessages($Share[SHARE_OBJ_ID],$Share[SHARE_TYPE]);
	}
	protected function OnBeforeDestroy(){
		parent::OnBeforeDestroy();
		$this->SaveShareList([],false);
	}
	
	// -----------------------------------------
	// -  Apply Maintain
	// -----------------------------------------
	protected function ApplyCleanUp(){
		parent::ApplyCleanUp();
		$clientList	= $this->LoadClientList();
		$changed	= false;
		foreach($clientList as &$client){
			if(!empty($client['LinkedIDs'])){
				$client['LinkedIDs']=[];
				$changed=true;
			}
		}
		if($changed)$this->SaveClientList($clientList);
// 				if($this->ReadPropertyBoolean('AutoSendShares'))$this->SendShares([]);
		
	}
	
	Protected function ApplyData(){
		parent::ApplyData();
		$shares=$this->LoadShareList();
		if(count($shares)==0)return $this->ApplyCleanUp();
		$rebuildIndex=false;
		// Check if client allowed State changed
		$this->UpdateClientsAllowed();
		
		// Remove Duplicates
		$myKeys = array_map(function($s){return (int)$s[SHARE_OBJ_ID];},$shares);
		$dupe=array_keys(array_flip($myKeys));
// $this->Debug("Keys: ".json_encode($myKeys));		
// $this->Debug("Dupe: ".json_encode($dupe));		
		if(count($dupe)!=count($myKeys)){
			foreach ($dupe as $objectID){
				$find=array_keys(array_filter($shares,function($i)use($objectID){return $i[SHARE_OBJ_ID]==$objectID;}));
				if(count($find)>1){
					for($j=0;$j<count($find);$j++){
			 			$del=$find[$j];
			 			if($shares[$del][SHARE_TYPE]==0){
							if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Error duplicate share! Remove => ".json_encode($shares[$del]),__FUNCTION__,__LINE__);
			 				unset($shares[$del]);
			 				$rebuildIndex=true;
			 			}
			 		}
			 	}
			}
		}
		// Check if Object Exists
		foreach($shares as $index=>$share){
			if($share[SHARE_OBJ_ID]==0|| !IPS_ObjectExists($share[SHARE_OBJ_ID])){
				if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Error object not exist! Remove share {$share[SHARE_OBJ_ID]}",__FUNCTION__,__LINE__);
				unset($shares[$index]);
				$rebuildIndex=true;
			}
		}
		// Fix for Display shares as List to remove subitems from deleted instance item
		$shareRoots = array_filter($shares,function($s){return $s[SHARE_PARENT]==0;});
		$roots	= array_map(function($s){return $s[SHARE_OBJ_ID];},$shareRoots);
		$delete = array_filter($shares,function($s)use($roots){return $s[SHARE_PARENT]>0 && !in_array($s[SHARE_PARENT], $roots);});
		foreach(array_keys($delete) as $index){
			if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Error object not exist! Remove share {$shares[$index][SHARE_OBJ_ID]}",__FUNCTION__,__LINE__);
			unset($shares[$index]);
			$rebuildIndex=true;
		}
		if(count($shares)==0){
			$this->SaveShareList([],false);
			return;
		}
		// Wenn gelöscht order hinzugefügt , dann dafür sorgen das dier index 0-x wieder stimmt
		if($rebuildIndex)$shares=array_values($shares);
		
		
		// Check if new shares			
		$addKeys 	= array_filter($shares,function($s){return $s[SHARE_TYPE]==0;});
		$addValues	= [];
		if(count($addKeys)>0){
			foreach($addKeys as $index=>$share){
				$objectID=$share[SHARE_OBJ_ID];
				if(!($newShare=$this->CreateObjectShares($objectID))){ // Fails create ref , delte and skip
					if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Error create share for => $objectID",__FUNCTION__,__LINE__);
					unset($shares[$index]);
					$rebuildIndex=true;
					continue;
				}
				if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Create shares for $objectID Data => ".json_encode($newShare),__FUNCTION__,__LINE__);
				$objectID=key($newShare);
				$rebuildIndex=true;
				$shares[$index][SHARE_OBJ_ID]=$objectID;
				$shares[$index][SHARE_TYPE]=$newShare[$objectID][SHARE_TYPE];
				if($newShare[$objectID][SHARE_TYPE]==2)$shares[$index][SHARE_DIR]=$newShare[$objectID][SHARE_DIR];
				$shares[$index][SHARE_VTYPE]=$newShare[$objectID][SHARE_VTYPE];
				unset($newShare[$objectID]);
				foreach($newShare as $subID=>$subShare){
					$addValues[$index][]=[
						SHARE_OBJ_ID=>$subID,
						SHARE_DIR=>$subShare[SHARE_DIR],
						SHARE_TYPE=>$subShare[SHARE_TYPE],
						SHARE_STATE=>STATE_READY,
						SHARE_PARENT=>$subShare[SHARE_PARENT],
						SHARE_VTYPE=>$subShare[SHARE_VTYPE]
					];
				}
			}
			// Inserting SubValues for new Instance Shares
			if(count($addValues)>0){
				foreach(array_reverse($addValues,true) as $index=>$values){
					if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Merge new shared Module variables at $index Data => ".json_encode($values),__FUNCTION__);
					array_splice($shares,$index+1,null,$values);
				}
				$rebuildIndex=true;
				$shares=array_values($shares);
			}
		}
		// Sort shares for Tree View
		if($rebuildIndex)$this->ReSortSharesByParent($shares);
		// Now detect changes from ShareIndex
		$shareIndex = $this->LoadShareIndex();
		$indexDiff=$this->GetSharesIndexDiff($shares,$shareIndex);
		if($this->HasDiffChanges($indexDiff)){
			$this->ApplySharesIndexDiff($indexDiff,$shareIndex);
			$this->SaveShareIndex($shareIndex,false);
 			$this->BlockApply();
 			$this->SaveShareList($shares,false);
 			$this->UnBlockApply();
 			$changesResponse=$this->ShareIndexDiffToShareChangesResponse($indexDiff);
 			if($this->HasDiffChanges($changesResponse)){
				if($this->ReadPropertyBoolean('AutoSendShares')){
 					$this->DoSendChangesResponse($changesResponse);
				}
  			}
		}
		
		
	}
	
	// -----------------------------------------
	// -  Private Server Maintain
	// -----------------------------------------
	private function ConfigureClient(string $JSONString){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$this->Debug($JSONString,__FUNCTION__);
		}
		if($cfg=json_decode($JSONString,true)){
			if(!empty($cfg['Client'])){
				$clientIP=$cfg['Client'];
				unset($cfg['Client']);
				foreach($cfg as $k=>$v){
					if(is_null($v))unset($cfg[$k]);
					elseif($k=='ControlInterval'){
						if($v==0)unset($cfg[$k]);
						elseif($v<0)$cfg[$k]=0;
					}
				}
				if(!empty($cfg)){
					$clientStates=$this->LoadClientStates();
					if(isset($clientStates[$clientIP])){
						$this->DoSendConfigResponse(json_encode($cfg),false,$clientIP,$clientStates[$clientIP]['Port']);
					}
				}elseif(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Noting to send to $clientIP => ".json_encode($cfg),__FUNCTION__);
			}elseif(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("No ClientIP set to send config => ".json_encode($cfg),__FUNCTION__);
		}elseif(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Invalid config => $JSONString",__FUNCTION__);
		
		
	}
	
	
	private function CreateObjectShares($ObjectID,$ParentID=0){
		$o=IPS_GetObject($ObjectID);
		$r=[];
		if($o['ObjectType']==OT_INSTANCE){
			$r[$ObjectID]=[SHARE_TYPE=>OT_INSTANCE,SHARE_VTYPE=>0,SHARE_PARENT=>0];
			foreach($o['ChildrenIDs'] as $id){
				if($v=$this->CreateObjectShares($id,$ObjectID))$r[key($v)]=array_shift($v);
			}
		}
		elseif($o['ObjectType']==OT_VARIABLE){
			if($v=IPS_GetVariable($ObjectID)){
				$r[$ObjectID]=[
						SHARE_TYPE=>OT_VARIABLE,
						SHARE_DIR=>$v['VariableAction'] || $v['VariableCustomAction'] ? DIR_ALL:DIR_STOC,
						SHARE_VTYPE=>$v['VariableType'],
						SHARE_PARENT=>$ParentID,
				];
			}
		}
		elseif($o['ObjectType']==OT_LINK){ 
		 	$r=$this->CreateObjectShares(IPS_GetLink($ObjectID)['TargetID'],$ParentID);
		}
		return $r;
	}
	
}
?>