<?php
if(!defined('KERNEL_VERSION'))define('KERNEL_VERSION',IPS_GetKernelVersion());
require_once 'RemoteIOConstants.inc';
// define('REMOTEIO_DEBUG',10);


class BaseModuleConstants extends IPSModule {
	// -----------------------------------------
	// -  Global Constants
	// -----------------------------------------
	protected const TYPE_NAMES =['Category','Module','Variable','Script','Event','Media','Link','Unknown'];
	protected const STATE_NAMES =['Ready','Connected','Error','Disabled','Creating...'];
	protected const DIR_NAMES  =["Disabled","Server to Client","Client to Server","Server <=> Client"];
	protected const CSTATE_NAMES = ['Disconnected','Connected','Verified','Not allowed','Fails','CleanUp','Destroy'];
	protected const YESNO_NAMES = ['No','Yes'];
	protected const MSG_NAMES = [OM_CHANGENAME=>'Name',OM_CHANGEICON=>'Icon',OM_CHANGEPOSITION=>'Position',OM_CHANGEDISABLED=>'Disabled',OM_CHANGEHIDDEN=>'Hidden',VM_CHANGEPROFILENAME=>'Profile',VM_UPDATE=>'Value',IM_CHANGESTATUS=>'Status'];
	
	protected function TranslatedErrorMessage($ErrorCode, $TextOnly=true, $DefaultText=''){
		$msg=$this->Translate(is_numeric($ErrorCode)?GetErrorMessage($ErrorCode,false,$DefaultText):$ErrorCode);
		return $TextOnly?$msg:"($ErrorCode) $msg";
	}
	protected function TranslatedTypeNames(int $Type=null){
		if(!is_null($Type)){
			$unknown=empty(self::TYPE_NAMES[$Type])?$this->Translate("Unknown Type")." $Type":'';
			return $unknown?$unknown:$this->Translate(self::TYPE_NAMES[$Type]);		
		}
		$names=[];foreach(self::TYPE_NAMES as $name)$names[]=$this->Translate($name);
		return $names;
		
	}
	protected function TranslatedStateNames(int $State=null){
		if(!is_null($State)){
			$unknown=empty(self::STATE_NAMES[$State])?$this->Translate("Unknown State")." $State":'';
			return $unknown?$unknown:$this->Translate(self::STATE_NAMES[$State]);		
		}
		$names=[];foreach(self::STATE_NAMES as $name)$names[]=$this->Translate($name);
		return $names;
		
	}
	protected function TranslatedDirNames(int $Direction=null){
		if(!is_null($Direction)){
			$unknown=empty(self::DIR_NAMES[$Direction])?$this->Translate("Unknown Direction")." $Direction":'';
			return  $unknown?$unknown:$this->Translate(self::DIR_NAMES[$Direction]);		
		}
		$names=[];foreach(self::DIR_NAMES as $name)$names[]=$this->Translate($name);
		return $names;
	}
	protected function TranslatedClientState(int $State=null){
		if(!is_null($State)){
			$unknown=empty(self::CSTATE_NAMES[$State])?$this->Translate("Unknown Client State")." $State":'';
			return  $unknown?$unknown:$this->Translate(self::CSTATE_NAMES[$State]);		
		}
		$names=[];foreach(self::CSTATE_NAMES as $name)$names[]=$this->Translate($name);
		return $names;
	}
	protected function TranslatedYesNo(bool $State=null){
		if(!is_null($State))return $this->Translate(self::YESNO_NAMES[(int)$State]);		
		$names=[];foreach(self::YESNO_NAMES as $name)$names[]=$this->Translate($name);
		return $names;
	}
	protected function TranslatedMessageName(int $MsgOption=null){
		if(!is_null($MsgOption)){
			$unknown=empty(self::MSG_NAMES[$MsgOption])?$this->Translate("Unknown Message name")." $MsgOption":'';
			return  $unknown?$unknown:$this->Translate(self::MSG_NAMES[$MsgOption]);		
		}
		$names=[];foreach(self::MSG_NAMES as $name)$names[]=$this->Translate($name);
		return $names;
	}
	
}
class BaseModuleTools extends BaseModuleConstants {
	public function __construct($InstanceID){
		parent::__construct($InstanceID);
		$dbl=0;
		if(!defined('REMOTEIO_DEBUG')){
			if(IPS_InstanceExists($InstanceID)){
				if($this->GetStatus()>101){
					$dbl=$this->ReadPropertyInteger('DebugLevel');
				}
				if($this->GetStatus()==103){
					IPS_LogMessage("RemoteIO", "before Destroy");
					$this->OnBeforeDestroy();
				}
				
			} else 
				IPS_LogMessage("RemoteIO", "Destroyed");
		
				
			define('REMOTEIO_DEBUG',$dbl);
		}
	}
	public function Create(){
		parent::Create();
		$this->RegisterPropertyInteger('DebugLevel', 1);
	}
	// -----------------------------------------
	// -  Module Global
	// -----------------------------------------
	protected function SendDebug($Message, $Data, $Format=0){
// 		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$Message=get_class($this).$Message;
		parent::SendDebug($Message, $Data, $Format);
	}
	protected function Debug($Message,$Function='',$Line='',$Class=''){
		$s=$Function?$Function:'';
		if($Line)$s.=$s?':'.$Line:$Line;
		if(empty($s)&&empty($Class))$Class=get_class($this);
		if($Class)$s=$s?"$Class::$s":$Class;
		$this->SendDebug($s,$Message, 0);		
	}
	
	// -----------------------------------------
	// -  Module Events
	// -----------------------------------------
	protected function OnBeforeDestroy(){
		IPS_LogMessage(IPS_GetName($this->InstanceID), "Execute");
	}
	protected function OnStatusChanged(int $OldStatus, int $NewStatus){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("$OldStatus => $NewStatus",__FUNCTION__,__LINE__, __CLASS__);
	}
	// -----------------------------------------
	// -  Module Helpers
	// -----------------------------------------
	/**
	 * @param mixed $v Value to Display
	 * @param int $TextMode Modes are 0=false/true 1=error/ok 2=no/yes
	 * @param string $Klammer if set to [ or ( then returns [yes/no] or (yes/no) 
 	 * @return string Formated Boolean String
	 */
	Protected function Bool2Str($v,int $TextMode=2,string $Klammer=null){ 
		static $text=[0=>['false','true'], 1=>['Error','Ok'], 2=>['No','Yes'] ]	;
		$r=$text[(int)$TextMode];
		$v=(bool)$v;
		$r=$r[(int)$v];
		if(!is_null($Klammer)){
			$a=$Klammer=='('?['(',')']:['[',']'];
			$r=sprintf("%s%s%s",$a[0],$r,$a[1]);
		}
		return $r;
	}	
	private $stateCache=null;
	/**
	 * @return integer The current Module Status
	 */
	protected function GetStatus(){
		return is_null($this->stateCache)?$this->stateCache=IPS_GetInstance($this->InstanceID)['InstanceStatus']:$this->stateCache;
	}
	/**
	 * {@inheritDoc}
	 * @see IPSModule::SetStatus()
	 */
	protected function SetStatus($Status){
		if(($OldStatus=$this->GetStatus())!=$Status){
			parent::SetStatus($this->stateCache=$Status);
			$this->OnStatusChanged($OldStatus, $Status);
		}
	}
	/**
	 * @return boolean True if module status = 102 otherwise false
	 */
	protected function CheckStatus(){
		return 	$this->GetStatus()==102;
	}
}
class BaseModuleTransport extends BaseModuleTools {
	protected $ConnectParentGuid='';
	protected $IsServer = false;
	
	public function Create(){
		parent::Create();
		$this->RegisterPropertyInteger('ControlInterval',$this->IsServer?2:1);
		$this->RegisterPropertyString('Identifier',($this->IsServer?'S':'C').substr(''.time(),4));
		$this->RegisterTimer('ControlTimer',0,"IPS_RequestAction($this->InstanceID,'CONTROL_TIMER',true);");
		if($this->ConnectParentGuid)$this->ConnectParent($this->ConnectParentGuid);
	}
	public function RequestAction($Ident, $Value){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(__FUNCTION__."($Ident, $Value)", '','',__CLASS__);
		if($Ident=='CONTROL_TIMER'){
			$this->OnControlTimer($Value);
		}
		else if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Error! Unknown Action: $Ident received! Value => $Value",__FUNCTION__);
	}
	public function ApplyChanges() {
		parent::ApplyChanges();
		if($this->GetStatus()==102)$this->RestartTimer();
	}
	
	public function ReceiveData($JSONString) {
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug($JSONString, __FUNCTION__,__LINE__,__CLASS__);
		if(!$this->CheckStatus()){
			if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Instance in Error state, skip data => $JSONString", __FUNCTION__);
			return;
		}
		if(!($data=json_decode($JSONString))){
			$msg=$this->Translate('Error! Empty or invalid data received');
			if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("$msg, skip data => $JSONString",__FUNCTION__);
			IPS_LogMessage(IPS_GetName($this->InstanceID),__FUNCTION__. "::$msg => ".$JSONString);
			$this->SetBuffer('RECEIVE_CACHE', '');
			return ;
		}
		if(!IPS_SemaphoreEnter('WaitForReceiveData'.$this->InstanceID, 3000)){
			$msg=$this->Translate('Error! Semaphore "WaitForReceiveData" enter fails');
			if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("$msg, skip data => $JSONString", __FUNCTION__);
			IPS_LogMessage(IPS_GetName($this->InstanceID),__FUNCTION__. "::$msg");
			return false;
		}
		$SemaphoreLeave=function(){
			IPS_SemaphoreLeave('WaitForReceiveData'.$this->InstanceID);
			restore_exception_handler();
		};
		$exception_handler=function ($e)use($SemaphoreLeave) {
			$msg=$e->getMessage();
			if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("CATCH ERROR => $msg", 'ReceiveData');
			IPS_LogMessage(IPS_GetName($this->InstanceID),'CATCH ERROR => '.wordwrap($msg,80));
			$SemaphoreLeave();
		};
		set_exception_handler($exception_handler);
		$cache=$this->GetBuffer('RECEIVE_CACHE');
		if(strpos($data->Buffer,BUFFER_SUFFIX)===false){
			if(strlen($data->Buffer)>0){
				$this->SetBuffer('RECEIVE_CACHE',$cache.$data->Buffer);
				if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug('Cache Buffer => '.$data->Buffer,__FUNCTION__);
				return $SemaphoreLeave();
			}
			else {
				$this->SetBuffer('RECEIVE_CACHE','');
				if(!empty($data->Type)){
					if($data->Type==1)$this->OnClientSocketConnected($data->ClientIP, $data->ClientPort);
					elseif($data->Type==2)$this->OnClientSocketDisConnected($data->ClientIP, $data->ClientPort);
					elseif(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Invalid data Received => $JSONString",__FUNCTION__);
				}
				else if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Invalid data Received => $JSONString",__FUNCTION__);
				return $SemaphoreLeave();
			}
		}
		if(!isset($data->ClientIP))$data->ClientIP='';
		if(!isset($data->ClientPort))$data->ClientPort=0;
		$buffers=explode(BUFFER_SUFFIX, $cache.$data->Buffer);
		foreach($buffers as $id=>$b){
			$len=strlen($b)-1;
			if(empty($b)||($b[$len]!='}'&&$b[$len]!=']')){
				$Rest=$buffers[$id];
				$this->SetBuffer('RECEIVE_CACHE',$cache.$Rest);
				if($Rest && REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug('Cache Buffer => '.$Rest,__FUNCTION__);
				break;
			}
			if($response=$this->Decode_Buffer($buffers[$id],true)){
				$this->MaintainData($response, (string)$data->ClientIP, (int)$data->ClientPort);			
			}elseif(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Error parsing Buffer => ".$buffers[$id],__FUNCTION__);
		}
		return $SemaphoreLeave();
	}
	private function MaintainData($Response, $ClientIP, $ClientPort ){
		list($type,$subType,$ack,$id,$data)= $Response;
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?"from $ClientIP:$ClientPort ":'';
			$this->Debug($m."Type:$type Sub:$subType ID: $id ACK:".$this->Bool2Str($ack,BOOL_YESNO)." Data => ".json_encode($data), __FUNCTION__);
		}
		if($this->OnBeforeReceive($type, $subType, $id, $ClientIP, $ClientPort)===false){
			if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("OnBeforeReceive returns false, skip data",__FUNCTION__);
			return;
		}
		switch($type){
			case DTYPE_ACK:$this->OnAckReceived($data, $id, $ClientIP, $ClientPort); break;
			case DTYPE_ERROR:$this->OnErrorReceived($data[0],$data[1],$id, $ClientIP, $ClientPort); break;
			case DTYPE_MESSAGE:$this->OnMessageReceived($data, $ack, $id, $ClientIP, $ClientPort);break;
			case DTYPE_PING: $this->OnPingReceived($data, $ack, $id, $ClientIP, $ClientPort);break;
			case DTYPE_PONG: $this->OnPongReceived($data, $ack, $id, $ClientIP, $ClientPort);break;	
			case DTYPE_REQUEST:$this->OnRequestReceived($subType, $data, $id,  $ClientIP, $ClientPort);break; 
			case DTYPE_RESPONSE:$this->OnResponseReceived($subType, $data, $ack, $id, $ClientIP, $ClientPort);break; 
			case DTYPE_DESTROY:$this->OnDestroyReceived($data, $ClientIP, $ClientPort, $id);break;
			default: $this->OnUnknownDataReceived($type, $subType, $data, $id, $ClientIP, $ClientPort);
		}
		$this->OnAfterReceive($type, $subType, $id, $ClientIP, $ClientPort);
	}
	// -----------------------------------------
	// -  De/Encode Maintain
	// -----------------------------------------
	protected function Decode_Buffer(string $Buffer,bool $AsArray){
		return json_decode($Buffer,$AsArray);		
	}
	protected function Encode_Buffer($Buffer, $ForDestroy=false){
		return json_encode($Buffer);
	}
	// -----------------------------------------
	// -  Timer Maintain
	// -----------------------------------------
	protected function OnStatusChanged($OldStatus, $NewStatus){
		parent::OnStatusChanged($OldStatus, $NewStatus);
		if($NewStatus==102)$this->RestartTimer();else $this->StopTimer();
	}
	protected function OnControlTimer($Mode){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(__FUNCTION__."($Mode)",'','',__CLASS__);
		elseif(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Timer executed",__FUNCTION__);
	}
	protected function RestartTimer(){
		$interval=$this->ReadPropertyInteger('ControlInterval');
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Interval => $interval",__FUNCTION__);
		if($interval<0)$interval=0;
		if($interval)
			$this->SetTimerInterval('ControlTimer', $interval*60*60*1000);
		else $this->SetTimerInterval('ControlTimer', 0);
	}	
	protected function StopTimer(){
		$this->SetTimerInterval('ControlTimer', 0);
	}
	// -----------------------------------------
	// -  General Receive Events
	// -----------------------------------------
	protected function OnClientSocketConnected(string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug($m,__FUNCTION__);
		}
		
	}
	protected function OnClientSocketDisConnected(string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug($m,__FUNCTION__);
		}
		
	}
	protected function OnBeforeReceive($Type,$SubType, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Type:$Type Sub:$SubType ID: $Id$m",__FUNCTION__);
		}
	}
	protected function OnAfterReceive($Type,$SubType, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Type:$Type Sub:$SubType ID: $Id$m",__FUNCTION__);
		}
	}
	protected function OnUnknownDataReceived($Type,$SubType,$Data, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Type:$Type Sub:$SubType ID: $Id$m => ".json_encode($Data),__FUNCTION_);
		}
	}
	protected function OnErrorReceived($ErrorCode,$ErrorData, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("ERROR:".$this->TranslatedErrorMessage($ErrorCode,false)." ID: $Id$m => ".json_encode($ErrorData),__FUNCTION__);
		}elseif(REMOTEIO_DEBUG>DEBUG_NONE){
			$this->Debug("ERROR: $ErrorCode ID: $Id$m => ".$this->TranslatedErrorMessage($ErrorCode),__FUNCTION__);
		}
	}
	protected function OnAckReceived($AckData, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("ID: $Id$m => ".json_encode($AckData),__FUNCTION__);
		}
	}
	protected function OnMessageReceived($Message, bool $Ack, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($Message),__FUNCTION__);
		}
	}
	protected function OnPingReceived($PingData, bool $Ack, int $Id, string $ClientIP, int $ClientPort ){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($PingData),__FUNCTION__);
		}
	}
	protected function OnPongReceived($PongData, bool $Ack, int $Id, string $ClientIP, int $ClientPort ){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($PongData),__FUNCTION__);
		}
	}
	protected function OnDestroyReceived($Identifier, string $ClientIP, int $ClientPort, int $Id){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $Identifier".$m,__FUNCTION__);
		}
		
	}
	
	// -----------------------------------------
	// -  Data Receive Events
	// -----------------------------------------
	protected function OnRequestReceived($RequestType, $Request,int $Id,string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Type:$RequestType ID: $Id$m => ".json_encode($Request),__FUNCTION__);
		}
	}
	protected function OnResponseReceived($ResponseType, $Response,bool $Ack, int  $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Type:$ResponseType Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($Response),__FUNCTION__);
		}
// 		if($Ack)$this->DoSendAck(true,$ClientIP,$ClientPort,$Id);
	}
	// -----------------------------------------
	// -  Data Send Events
	// -----------------------------------------
	/**
	 * You can change in this event the ID , ClientIP or ClientPort. For multiple calls change in this Event then Client IP and Port to Arrays
	 * @param string $Type The general DTYPE of Data 
	 * @param string $SubType The RTYPE of Request or Response
	 * @param int $Id UserID (not used at tis time)
	 * @param string|string[] $ClientIP Ip or List of IPs to send
	 * @param int|int[] $ClientPort Port or list of ports to send . If ClientIP is array then every ClientIP must have a corespondening Port in this list
	 * @returns null|bool if abort sending you must return false
	 */
	protected function OnBeforeSend(string $Type,string $SubType, int &$Id, string &$ClientIP, int &$ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Type:$Type Sub:$SubType ID: $Id$m",__FUNCTION__);
		}
	}
	protected function OnAfterSend(string $Type,string $SubType, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Type:$Type Sub:$SubType ID: $Id$m",__FUNCTION__);
		}
	}
	// -----------------------------------------
	// -  General Sending
	// -----------------------------------------
	/**
	 * @param string $Type
	 * @param string $SubType
	 * @param mixed $Data
	 * @param bool $Ack
	 * @param string $ClientIP
	 * @param number $ClientPort
	 * @param int $Id
	 */
	protected function DoSendData(string $Type,string $SubType, $Data, bool $Ack=null, $ClientIP='',$ClientPort=0, $Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			if(is_array($ClientIP))$m=" to Clients ".json_encode($ClientIP);
			else $m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Type: $Type$m =>".json_encode($Data),__FUNCTION__);
		}
// 		if($Type==DTY)
		if(!$this->CheckStatus()){
			if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Instance has Error status ".$this->GetStatus().", skip send",__FUNCTION__);
			return;
		}
		
		if($this->OnBeforeSend($Type, $SubType,$Id,$ClientIP,$ClientPort)===false){
			if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("OnBeforeSend returns false, skip send", __FUNCTION__);
			return;
		}
		
		$ips=is_array($ClientIP)?$ClientIP:[$ClientIP];
		$ports=is_array($ClientPort)?$ClientPort:[$ClientPort];
		$buffer=$this->Encode_Buffer([
				$Type, // DataType
				$SubType, // Response / Request Type
				(bool)$Ack, // ACK Request
				$Id, // ID
				$Data
		]).BUFFER_SUFFIX;
		$bufLen=strlen($buffer);
		
		foreach($ips as $index=>$ClientIP){
			if(!isset($ports[$index]))continue;
			$ClientPort=$ports[$index];
			$send=[
				'DataID'=>$ClientIP?'{C8792760-65CF-4C53-B5C7-A30FCC84FEFE}':'{79827379-F36E-4ADA-8A95-5F8D1DC92FA9}',
				'ClientIP'=>$ClientIP,
				'ClientPort'=>$ClientPort,
				'Buffer'=>$buffer
			];
			$send=json_encode($send);
			if(REMOTEIO_DEBUG>DEBUG_DETAIL){
				$dataLen=strlen($send);
				$this->Debug("Send BufSize=$bufLen, DataSize=$dataLen Data =>".$send,__FUNCTION__);
			}
			$this->SendDataToParent($send);
			$this->OnAfterSend($Type, $SubType, $Id,$ClientIP,$ClientPort);
		}
	}
	protected function DoSendDestroy($ClientIP='',$ClientPort=0, $Id=0){
 		if(empty($Id))$Id=$this->IsServer?-1:1;
		$buffer=$this->Encode_Buffer([DTYPE_DESTROY, 0,false,$Id,$this->ReadPropertyString('Identifier')]).BUFFER_SUFFIX;
		$this->SendDataToParent(json_encode([
			'DataID'=>$ClientIP?'{C8792760-65CF-4C53-B5C7-A30FCC84FEFE}':'{79827379-F36E-4ADA-8A95-5F8D1DC92FA9}',
			'ClientIP'=>$ClientIP,
			'ClientPort'=>$ClientPort,
			'Buffer'=>$buffer
		]));
		
	}
	protected function DoSendError($ErrorCode, $ErrorData, $ClientIP='',$ClientPort=0, $Id=0 ){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Code:$ErrorCode ID: $Id$m => ".json_encode($ErrorData),__FUNCTION__);
		}
		$this->DoSendData(DTYPE_ERROR, '', [$ErrorCode, $ErrorData], false,$ClientIP,$ClientPort, $Id=0);
	}
	protected function DoSendAck($AckData=null, $ClientIP='',$ClientPort=0, $Id=0 ){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("ID: $Id$m => ".json_encode($AckData),__FUNCTION__);
		}
		$this->DoSendData(DTYPE_ACK, '', $AckData, false,$ClientIP,$ClientPort,$Id);
	}
	protected function DoSendMessage(array $Message, $Ack=false, $ClientIP='', $ClientPort=0, $Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($Message),__FUNCTION__);
		}
		$this->DoSendData(DTYPE_MESSAGE, '', $Message, $Ack,$ClientIP,$ClientPort,$Id);
	}
	protected function DoSendPing($PingData=null, $Ack=false, $ClientIP='', $ClientPort=0, $Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($PingData),__FUNCTION__);
		}
		$this->DoSendData(DTYPE_PING, '', $PingData, $Ack,$ClientIP,$ClientPort,$Id);
	}
	protected function DoSendPong($PongData=null, $Ack=false, $ClientIP='', $ClientPort=0, $Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($PongData),__FUNCTION__);
		}
		$this->DoSendData(DTYPE_PONG, '', $PongData, $Ack,$ClientIP,$ClientPort,$Id);
	}
	// -----------------------------------------
	// -  Data Sending Request/Response
	// -----------------------------------------
	protected function DoSendRequest(string $RequestType, $RequestData,  $ClientIP='',$ClientPort=0 , $Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Type:$RequestType ID: $Id$m => ".json_encode($RequestData),__FUNCTION__);
		}
		$this->DoSendData(DTYPE_REQUEST, $RequestType, $RequestData, false,$ClientIP,$ClientPort ,$Id=0);
	}
	protected function DoSendResponse($ResponseType, $ResponseData, $Ack=false, $ClientIP='',$ClientPort=0 ,$Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Type:$ResponseType Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($ResponseData),__FUNCTION__);
		}
		$this->DoSendData(DTYPE_RESPONSE, $ResponseType, $ResponseData, $Ack,$ClientIP,$ClientPort,$Id);
	}
}
class BaseModuleObjects extends BaseModuleTransport {
	// -----------------------------------------
	// -  Data Maintain
	// -----------------------------------------
	protected function GetLocaleIDforRemoteID(int $RemoteID){
		return $RemoteID;
	}
	protected function GetRemoteIDforLokaleID(int $LocaleID){
		return $LocaleID;
	}
	// -----------------------------------------
	// -   Objects Events
	// -----------------------------------------
	protected function OnLocalObjectStateChanged(int $LocaleID, int $State){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$this->Debug("LocaleID: $LocaleID State: $State",__FUNCTION__);
		}
	}
	// -----------------------------------------
	// -   Objects Data Events
	// -----------------------------------------
	protected function OnRequestReceived($RequestType, $Request,int $Id, string $ClientIP, int $ClientPort){
		parent::OnRequestReceived($RequestType, $Request, $Id, $ClientIP, $ClientPort);
		if($RequestType==RTYPE_OBJECT ){
			$this->OnObjectRequest($Request[0],$Request[1], $ClientIP, $ClientPort, $Id);
		}
		elseif($RequestType==RTYPE_PROFILE ){
			$this->OnProfileRequest($Request, $ClientIP, $ClientPort, $Id);
		}
	}
	protected function OnResponseReceived($ResponseType, $Response, bool $Ack, int $Id, string $ClientIP, int $ClientPort){
		parent::OnResponseReceived($ResponseType, $Response, $Ack, $Id, $ClientIP, $ClientPort);
		if($ResponseType==RTYPE_OBJECT ){
			$this->OnObjectResponse($Response[0],$Response[1],$Ack, $ClientIP, $ClientPort,$Id);
		}
		elseif($ResponseType==RTYPE_PROFILE ){
			$this->OnProfileResponse($Response, $Ack, $ClientIP, $ClientPort, $Id);
		}
	}
	// -----------------------------------------
	// -  Objects Data Request/Response
	// -----------------------------------------
	protected function OnObjectRequest(array $ObjectIDs, int $FilterOptions, string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort ":'';
			$this->Debug($m."Filter:$FilterOptions => ".json_encode($ObjectIDs),__FUNCTION__);
		}
	}
	protected function OnObjectResponse(array $ObjectDatas, int $FilterOptions, bool $Ack, string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Filter:$FilterOptions Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($ObjectDatas),__FUNCTION__);
		}
	}
	/**
	 * Its called befor change Object Value specified with $ValueOption
	 * @param int $LocaleID
	 * @param int $ValueOption
	 * @param mixed $NewValue
	 * @returns NULL|mixed If not null result will be uses as return value and process finishd  
	 */
	protected function OnBeforeObjectValueSet(int $LocaleID, int $ValueOption, $NewValue){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$this->Debug("LocaleID: $LocaleID ValueOption: $ValueOption NewValue => ".json_encode($NewValue),__FUNCTION__);
		}
	}
	protected function OnBeforeLocalDataChanged(int $LocaleID, int $RemoteID, int $MessageID, $NewValue){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			if(is_bool($NewValue))$NewValue=$this->Bool2Str($NewValue);
			$this->Debug("Sender:$LocaleID RemoteID: $RemoteID MessageID: $MessageID NewValues: $NewValue",__FUNCTION__);
		}
	}
	/**
	 * This will be called after local data on object changed, you must enable here Messagesinks for Object
	 * @param int $LocaleID
	 * @param int $RemoteID
	 * @param int $MessageID
	 * @param bool $Modified
	 */
	protected function OnAfterLocalDataChanged(int $LocaleID, int $RemoteID, int $MessageID, bool $Modified){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$this->Debug("Sender:$LocaleID RemoteID: $RemoteID MessageID: $MessageID Modified: ".$this->Bool2Str($Modified,2),__FUNCTION__);
		}
	}

	
	// -----------------------------------------
	// -  Objects Profile Request/Response
	// -----------------------------------------
	protected function OnProfileRequest(array $ObjectIDs,string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort ":'';
			$this->Debug($m."ID: $Id => ".json_encode($ObjectIDs),__FUNCTION__);
		}
	}
	protected function OnProfileResponse(array $ProfileDatas,bool $Ack,string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($ProfileDatas),__FUNCTION__);
		}
		
	}
	protected function OnBeforeLocalProfileChanged(int $LocaleID, string $NewProfile){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$this->Debug("LocaleID: $LocaleID NewProfile => $NewProfile",__FUNCTION__);
		}
	}
	
	protected function OnAfterLocalProfileChanged(int $LocaleID, string $NewProfile, bool $Modified){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$this->Debug("LocaleID: $LocaleID Modified: ".$this->Bool2Str($Modified,2)." Profile => $NewProfile",__FUNCTION__);
		}
	}
	// -----------------------------------------
	// -  Objects Data Sending
	// -----------------------------------------
	protected function DoSendObjectRequest(array $ObjectIDs, int $FilterOptions, $ClientIP='',$ClientPort=0 ,$Id=0, $UserData=null){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Filter:$FilterOptions ID: $Id$m => ".json_encode($ObjectIDs),__FUNCTION__);
		}
		$this->DoSendRequest(RTYPE_OBJECT, [$ObjectIDs,$FilterOptions], $ClientIP,$ClientPort,$Id,$UserData);
	}
	protected function DoSendObjectResponse(array $ObjectData, int $FilterOptions, $Ack=false, $ClientIP='',$ClientPort=0 ,$Id=0, $UserData=null){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Filter:$FilterOptions Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($ObjectData),__FUNCTION__);
		}
		$this->DoSendResponse(RTYPE_OBJECT, [$ObjectData, $FilterOptions], $Ack, $ClientIP,$ClientPort,$Id, $UserData);
	}
	protected function DoSendProfileRequest(array $ObjectIDs, $ClientIP='',$ClientPort=0 ,$Id=0, $UserData=null){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("ID: $Id$m => ".json_encode($ObjectIDs),__FUNCTION__);
		}
		$this->DoSendRequest(RTYPE_PROFILE, $ObjectIDs, $ClientIP,$ClientPort,$Id,$UserData);
	}
	protected function DoSendProfileResponse(array $ProfileDatas, $Ack=false, $ClientIP='',$ClientPort=0 ,$Id=0, $UserData=null){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($ProfileDatas),__FUNCTION__);
		}
		$this->DoSendResponse(RTYPE_PROFILE, $ProfileDatas, $Ack, $ClientIP,$ClientPort,$Id, $UserData);
	}
	
	// -----------------------------------------
	// -  Objects Profile Maintain
	// -----------------------------------------
	protected function GetProfileData(int $LocaleID){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("for $LocaleID",__FUNCTION__);
		if($v=@IPS_GetVariable($LocaleID)){
			if(!($p=$v['VariableCustomProfile']))$p=$v['VariableProfile'];
			
			if(!empty($p)){
				if($p[0]!='~'){
					if(IPS_VariableProfileExists($p)){
						$p=IPS_GetVariableProfile($p);
						foreach($p['Associations'] as &$a)$a=array_values($a);
						if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("for $LocaleID returns => ".json_encode(array_values($p)),__FUNCTION__);
						return array_values($p);
					}else/*DBG*/if(REMOTEIO_DEBUG>0)$this->Debug("Error! Profile $p for $LocaleID not exist",__FUNCTION__);
				}else/*DBG*/if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Standard Profile $p for Variable $LocaleID not copyed,skip",__FUNCTION__);
			}else/*DBG*/if(REMOTEIO_DEBUG>0)$this->Debug("Variable $LocaleID has no Profile",__FUNCTION__);
		}else/*DBG*/if(REMOTEIO_DEBUG>0)$this->Debug("Variable $LocaleID not exist",__FUNCTION__);
	}
	protected function SetProfileData(int $LocaleID, array $ProfileData){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("LocaleID: $LocaleID Data => ".json_encode($ProfileData),__FUNCTION__);
		if(IPS_VariableExists($LocaleID)){
			if($profile=$this->CreateProfileFromData($ProfileData)){
				if($this->OnBeforeLocalProfileChanged($LocaleID, $profile)!==false){
					$ok=IPS_SetVariableCustomProfile($LocaleID, $profile);
					$this->OnAfterLocalProfileChanged($LocaleID, $profile, $ok);
					return $ok;
				}
				
			}
		}
		elseif(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("VariableID: $LocaleID not exist",__FUNCTION__);
			
		
	}
	protected function CreateProfileFromData(array $ProfileData){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug( json_encode($ProfileData),__FUNCTION__);
		try {
			$ProfileData=array_combine(array_keys(IPS_GetVariableProfile('~Switch')), $ProfileData);
		}catch(Exception $e){
			if(REMOTEIO_DEBUG>DEBUG_NONE)$this-Debug("Invalid profiledata => ".json_encode($ProfileData),__FUNCTION__);
			return false;
		}
		$pname=$ProfileData['ProfileName'];
		if($pname[0]=='~')return $pname;
		if(!IPS_VariableProfileExists($pname))$pname='REMOTE_'.$pname;
		if(!IPS_VariableProfileExists($pname)){
			IPS_CreateVariableProfile($pname, $ProfileData['ProfileType']);
			if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Profile ->$pname<- created",__FUNCTION__);
		} else {
			if(IPS_GetVariableProfile($pname)['ProfileType']!=$ProfileData['ProfileType']){
				if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Profile ->$pname<- exist but Type not match",__FUNCTION__);
				return false;
			}
		}
		IPS_SetVariableProfileDigits($pname, $ProfileData['Digits']);
		IPS_SetVariableProfileIcon($pname, $ProfileData['Icon']);
		IPS_SetVariableProfileText($pname, $ProfileData['Prefix'], $ProfileData['Suffix']);
		IPS_SetVariableProfileValues($pname, $ProfileData['MinValue'], $ProfileData['MaxValue'], $ProfileData['StepSize']);
		foreach($ProfileData['Associations'] as $a){
			IPS_SetVariableProfileAssociation($pname, $a[0], $a[1], $a[2], $a[3]);
		}
		return $pname;
	}

	// -----------------------------------------
	// -  Objects Data Maintain
	// -----------------------------------------
	protected function GetObjectData(int $LocaleID, int $ValueOptions){
		if($ValueOptions<1 || !IPS_ObjectExists($LocaleID))return null;
		$object	= IPS_GetObject($LocaleID);
		$return	= [];
		if($ValueOptions & VOPT_NAME)$return[VOPT_NAME]=$object['ObjectName'];
		if($ValueOptions & VOPT_ICON)$return[VOPT_ICON]=$object['ObjectIcon'];
		if($ValueOptions & VOPT_POSITION)$return[VOPT_POSITION]=$object['ObjectPosition'];
		if($ValueOptions & VOPT_DISABLED)$return[VOPT_DISABLED]=$object['ObjectIsDisabled'];
		if($ValueOptions & VOPT_IDENT)$return[VOPT_IDENT]=$object['ObjectIdent'];
		if($ValueOptions & VOPT_HIDDEN)$return[VOPT_HIDDEN]=$object['ObjectIsHidden'];
		if(($ValueOptions & (VOPT_PROFILE+VOPT_VALUE)) && $object['ObjectType']== OT_VARIABLE ){
			if($ValueOptions & VOPT_PROFILE){
				$v=IPS_GetVariable($LocaleID);
				$p=empty($v['VariableCustomProfile'])?$v['VariableProfile']:$v['VariableCustomProfile'];
				$return[VOPT_PROFILE]=$p;
			}
			if($ValueOptions& VOPT_VALUE)$return[VOPT_VALUE]=GetValue($LocaleID);
		} else {
			if($ValueOptions & VOPT_PROFILE)$return[VOPT_PROFILE]=null;
			if($ValueOptions & VOPT_VALUE)$return[VOPT_VALUE]=null;
		}
		if($ValueOptions & VOPT_STATUS && $object['ObjectType']== OT_INSTANCE){
			if($i=@IPS_GetInstance($LocaleID))
				$return[VOPT_STATUS]=$i['InstanceStatus'];
			else $return[VOPT_STATUS]=null;
		}
		return count($return)==1?array_pop($return):$return;
	}

	/**
	 * @param int $LocaleID
	 * @param mixed $Values
	 * @param int $ValueOptions
	 * @param bool $ChangesOnly
	 * @return array|NULL|mixed
	 */
	protected function SetObjectData(int $LocaleID, $Values, int $ValueOptions, bool $ChangesOnly=true){
		if($LocaleID>0 && !IPS_ObjectExists($LocaleID))$LocaleID=0;
		$filtered= is_array($Values);
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("LocaleID: $LocaleID Filtered: ".$this->Bool2Str($filtered). " Options: $ValueOptions Values => ".json_encode($Values),__FUNCTION__);
		$return=$filtered?[]:[true];
		$profileChanged=null;
		$newProfile='';
		for($j=0;$j<10;$j++){
			$option=1<<$j;
			$handled=null;
			if($ValueOptions==0 || ( ( $filtered && $ValueOptions & $option) || (!$filtered && $ValueOptions==$option) ) ){
				$value=$filtered?@$Values[$option]:$Values;
				if($option==VOPT_PROFILE && !is_null($value)){
					$profileChanged=false;$newProfile=$value;
				}
				if(is_null($value) || !is_null($handled=$this->OnBeforeObjectValueSet($LocaleID, $option, $value)) || $LocaleID==0){
					if($filtered)$return[$option]=$handled;
					continue;
				}
				if(!$ChangesOnly || $value!=$this->GetObjectData($LocaleID, $option) ){
					$ok=false;
					switch($option){
						case VOPT_NAME		: $ok=IPS_SetName($LocaleID, $value);break;
						case VOPT_ICON		: $ok=IPS_SetIcon($LocaleID, $value);break;
						case VOPT_POSITION	: $ok=IPS_SetPosition($LocaleID, $value) ;break;
						case VOPT_DISABLED	: $ok=IPS_SetDisabled($LocaleID, $value);break;
						case VOPT_HIDDEN	: $ok=IPS_SetHidden($LocaleID, $value);break;
						case VOPT_IDENT		: $ok=IPS_SetIdent($LocaleID, $value);break;
						case VOPT_VALUE		: $ok=@SetValue($LocaleID, $value);break;
						case VOPT_PROFILE	: $ok=$profileChanged=IPS_SetVariableCustomProfile($LocaleID, $value);break;
						case VOPT_STATUS	: $ok=($this->OnLocalObjectStateChanged($LocaleID, $value)===false);break;
					}
					if(!$ok){
						$value=[false,ERR_CantChangeObjectProps, null];
					}else $value=[true];
					if($filtered)$return[$option]=$value; else $return=$value;
				} elseif(!$filtered)$return=[true];
			}
		}
		if(!is_null($profileChanged)){
			$this->OnAfterLocalProfileChanged($LocaleID, $newProfile , $profileChanged);
		}
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)	$this->Debug("Returns => ".json_encode($return),__FUNCTION__,__LINE__);
		return $return;
	}
	
}

class BaseModuleMessages extends BaseModuleObjects{
	public function Create(){
		$this->RegisterDefaultMessages();
		parent::Create();
		$this->RegisterPropertyBoolean('SendChangesOnly', true);
// 		$this->RegisterMessage(0, IPS_INSTANCEMESSAGE);
	}
	public function MessageSink($TimeStamp, $SenderID, $Message, $Data) {
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("SenderID $SenderID,Message $Message Data ".json_encode($Data), __FUNCTION__);
		
		
		if($Message==FM_CONNECT){
 			$i=IPS_GetInstance($Data[0]);
 			$t=IPS_GetName($Data[0]);
			if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Parent instance \"$t\" connected => ".$Data[0],__FUNCTION__);
 			if($i['ModuleInfo']['ModuleID'] == $this->ConnectParentGuid){
 				$this->RegisterMessage($Data[0], IM_CHANGESTATUS);
				$this->SetBuffer('LAST_CONNECTION_ID',$Data[0]);
				$this->SetStatus($i['InstanceStatus']);
			} else $this->SetStatus(201);
 			return ;
 		}
 		elseif($Message==FM_DISCONNECT){
 			$lastID=intval($this->GetBuffer('LAST_CONNECTION_ID'));
 			if($lastID){
				$this->UnRegisterMessage($lastID, IM_CHANGESTATUS);
 			}
			$this->SetBuffer('LAST_CONNECTION_ID',0);
			if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Parent instance disconnected => ".$lastID,__FUNCTION__);
 			$this->SetStatus(200);
			return ;
		}
		elseif($Message==IM_CHANGESTATUS){
			$MyParentID=intval($this->GetBuffer('LAST_CONNECTION_ID'));
			if($SenderID!=$this->InstanceID && $SenderID!=$MyParentID){
				$this->OnSharedInstanceStateChanged($SenderID, (int)$Data[0]);
				return;
			}
			elseif($SenderID==$this->InstanceID){
				IPS_LogMessage(get_class($this), "Status : ".$Data[0]);
				return;
			} 
			if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Parent instance status change to => ".$Data[0],__FUNCTION__);
			$this->SetStatus($Data[0]);
			
		}
		elseif($Message == IPS_KERNELMESSAGE){
			try {
				if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Kernel Message => ".$Data[0],__FUNCTION__);
				switch($Data[0]){
					case  KR_READY : $this->OnKernelReady();IPS_Sleep(150);break;
					case  KR_UNINIT: $this->OnKernelDown();break;
					case  KR_SHUTDOWN: $this->OnKernelDown();break;
					
				}
				
			}catch(Exception $e){
				$this->Debug("CATCH ERROR => ".$e->getMessage(),__FUNCTION__);
			}
			return;
		}
		else $this->MaintainMessageSink((int)$SenderID, (int)$Message, isset($Data[0])?$Data[0]:null, isset($Data[1])?$Data[1]:null,isset($Data[2])?$Data[2]:null);
	}
	private function MaintainMessageSink(int $LocaleID,int $Message,$NewValue,$Data1,$OldValue){
		if($this->OnMessageSink($LocaleID, $Message, $NewValue, $Data1, $OldValue)===true)return true;
		if($Message==VM_DELETE)	$this->OnLocalVariableDeleted($LocaleID);
		elseif($Message==IM_DELETE)$this->OnLocalInstanceDeleted($LocaleID);
		elseif(!($RemoteID=$this->GetRemoteIDforLokaleID($LocaleID))){
			if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("No RemoteID for LocalID $LocaleID found",__FUNCTION__);
		}
		else if($this->ReadPropertyBoolean('SendChangesOnly') && $NewValue==$OldValue){
			if(REMOTEIO_DEBUG>DEBUG_ERROR){
				if(is_bool($NewValue))$NewValue=$this->Bool2Str($NewValue,0);
				$this->Debug("Skip send for $RemoteID, SendChangesOnly is set and old and new values are the same => ".$NewValue,__FUNCTION__);
			}
		}
		else $this->DoSendMessage([$RemoteID, $Message, $NewValue,$Data1,$OldValue]);
	}
	// -----------------------------------------
	// -  Override Events Maintain
	// -----------------------------------------
	protected function OnStatusChanged($OldStatus, $NewStatus){
		if($OldStatus==102)
			$this->OnInstanceError();
		elseif($NewStatus==102)
			$this->OnInstanceReady();
	}
	protected function OnMessageReceived($Message, bool $Ack, int $Id, string $ClientIP, int $ClientPort){
		parent::OnMessageReceived($Message, $Ack, $Id, $ClientIP, $ClientPort);
		$return=[false,0,null];
		$updateEvents=true;
		list($remoteID, $messageID, $newValue)=$Message;
		$convert=[
				OM_CHANGENAME=>VOPT_NAME,
				OM_CHANGEICON=>VOPT_ICON,
				OM_CHANGEPOSITION=>VOPT_POSITION,
				OM_CHANGEDISABLED=>VOPT_DISABLED,
				OM_CHANGEHIDDEN=>VOPT_HIDDEN,
				VM_CHANGEPROFILENAME=>VOPT_PROFILE,
				VM_UPDATE=>VOPT_VALUE,
				IM_CHANGESTATUS=>VOPT_STATUS
		];
		$msgOption=isset($convert[$messageID])?$convert[$messageID]:0;
		if($msgOption==0){
			$return=[false,ERR_MessageNoSupported,null];
		}
		elseif(!$localeID=$this->GetLocaleIDforRemoteID($remoteID)){
			$return=[false,ERR_LocalForRemoteIDExist,null];
		}
		elseif(($myValue=$this->GetObjectData($localeID, $msgOption))===null){
			$return=[false,ERR_MessageNoSupported,null];
		}
		elseif($myValue==$newValue){
			$updateEvents=false;
			$return=[true];
		}
		elseif($this->OnBeforeLocalDataChanged($localeID, $remoteID, $messageID, $newValue)===false){
			$return=[false,ERR_BeforeLocalDataChanged,null];
		}
		else { 
			$return=$this->SetObjectData($localeID, $newValue, $msgOption);
			if(REMOTEIO_DEBUG>DEBUG_NONE && $return[0]===false ){
				$this->Debug("ID: $localeID (".$this->TranslatedMessageName($messageID).") Error => ".json_encode($return), __FUNCTION__);
			}elseif(REMOTEIO_DEBUG>DEBUG_ERROR){
				$this->Debug("ID: $localeID (".$this->TranslatedMessageName($messageID).") Changed: ". $this->Bool2Str($return[0])." to => $newValue", __FUNCTION__);
			}
		}
		if($updateEvents && $localeID>0)$this->OnAfterLocalDataChanged($localeID, $remoteID, $messageID,$return[0]);
	
		if($return[0]===false){
 			$this->DoSendError($return[1],$Message,$ClientIP,$ClientPort,$Id);
		}elseif($Ack){
			$this->DoSendMessage($Message,false,$ClientIP,$ClientPort,$Id);
		}
		
	}
	// -----------------------------------------
	// -  Local MessageSink Maintain
	// -----------------------------------------
	protected function OnMessageSink(int $LocaleID,int $MessageID, $NewValue,$Data1,$OldValue){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			if(is_bool($NewValue))$NewValue=$this->Bool2Str($NewValue);
			if(is_bool($OldValue))$OldValue=$this->Bool2Str($OldValue);
			$this->Debug("Sender:$LocaleID Message:$MessageID Values: $OldValue => $NewValue",__FUNCTION__);
		}
	}
	// -----------------------------------------
	// -  System Events Maintain
	// -----------------------------------------
	protected function OnKernelReady(){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("-------------- Kernel Ready ----------------",__FUNCTION__);
	}
	protected function OnKernelDown(){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Kernel goes Down",__FUNCTION__);
	}
	protected function OnInstanceReady(){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Connection ready",__FUNCTION__);
	}
	protected function OnInstanceError(){
		if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Connection closed", __FUNCTION__);
	}	
	// -----------------------------------------
	// -  State Events Maintain
	// -----------------------------------------
	protected function OnSharedInstanceStateChanged(int $ObjectID, int $NewState){
		if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("ObjectID: $ObjectID Status => $NewState",__FUNCTION__);
	}
	// -----------------------------------------
	// -   Data Events
	// -----------------------------------------
	protected function OnLocalVariableDeleted(int $VariableID){
		if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Local Variable $VariableID Deleted",__FUNCTION__);
	}
	protected function OnLocalInstanceDeleted(int $InstanceID){
		if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Local Instance $InstanceID Deleted",__FUNCTION__);
	}
	// -----------------------------------------
	// -   Messages Maintain
	// -----------------------------------------
	private $messages=[
		0=>[OM_CHANGENAME,OM_CHANGEICON,OM_CHANGEPOSITION,OM_CHANGEDISABLED], // Object default;
		1=>[IM_CHANGESTATUS], // Instanz
		2=>[VM_UPDATE, VM_CHANGEPROFILENAME], // Variable
		3=>[], //[SM_DELETE, SM_CHANGEFILE,SM_BROKEN], // Skript
		4=>[], //[EM_DELETE, EM_CHANGEACTIVE], // Ereignis
		5=>[], //[MM_DELETE, MM_UPDATE,MM_CHANGEFILE,MM_CHANGECACHED], // Media,
		6=>[]  //[LM_DELETE, LM_CHANGETARGET]  // Link
	]; 
	protected function RegisterDefaultMessages(){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("()",__FUNCTION__,'',get_class($this));
		$this->RegisterMessage($this->InstanceID, FM_CONNECT);
		$this->RegisterMessage($this->InstanceID, FM_DISCONNECT);
 		$this->RegisterMessage(0, IPS_KERNELMESSAGE);
	}
	protected function GetObjectMessages(int $LocaleID, $ObjectType=null){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("LocaleID: $LocaleID ObjectType: ".$this->TranslatedTypeNames($ObjectType),__FUNCTION__);
		if(is_null($ObjectType))$ObjectType=IPS_GetObject($LocaleID)['ObjectType'];
		if($ObjectType>0||$ObjectType<7){
			$messages=$this->messages[0];
			if($ObjectType>0)$messages=array_merge($messages,$this->messages[$ObjectType]);
		} else $messages=[];
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Returns => ".json_encode($messages),__FUNCTION__);
		return $messages;
	}
	protected function RegisterObjectMessages(int $LocaleID, $ObjectType=null){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("LocaleID: $LocaleID ObjectType: ".$this->TranslatedTypeNames($ObjectType),__FUNCTION__);
		foreach($this->GetObjectMessages($LocaleID, $ObjectType) as $msgid){
			$this->RegisterMessage($LocaleID, $msgid);
		}
	}
	protected function UnRegisterObjectMessages(int $LocaleID, $ObjectType=null){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("LocaleID: $LocaleID ObjectType: ".$this->TranslatedTypeNames($ObjectType),__FUNCTION__);		
		foreach($this->GetObjectMessages($LocaleID, $ObjectType) as $msgid){
			$this->UnRegisterMessage($LocaleID, $msgid);
		}
	}
	protected function RegisterObject(int $LocaleID,$ObjectType=null){
		if(is_null($ObjectType))$ObjectType=IPS_GetObject($LocaleID)['ObjectType'];
		if(empty($ObjectType))return;
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("LocaleID: $LocaleID ObjectType: ".$this->TranslatedTypeNames($ObjectType),__FUNCTION__);		
		$table=[1=>IM_DELETE,2=>VM_DELETE,3=>EM_DELETE,4=>SM_DELETE,5=>MM_DELETE,6=>LM_DELETE];
		$this->RegisterMessage($LocaleID, $table[$ObjectType]);
		if(KERNEL_VERSION>5)$this->RegisterReference($LocaleID);
	}
	protected function UnRegisterObject(int $LocaleID,$ObjectType=null){
		if(is_null($ObjectType))$ObjectType=IPS_GetObject($LocaleID)['ObjectType'];
		if(empty($ObjectType))return;
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("LocaleID: $LocaleID ObjectType: ".$this->TranslatedTypeNames($ObjectType),__FUNCTION__);		
		$table=[1=>IM_DELETE,2=>VM_DELETE,3=>EM_DELETE,4=>SM_DELETE,5=>MM_DELETE,6=>LM_DELETE];
		$this->UnRegisterMessage($LocaleID, $table[$ObjectType]);
		if(KERNEL_VERSION>5)$this->UnRegisterReference($LocaleID);
	}
	
	protected function MaintainObjectMessages(int $LocaleID, bool $Enable, int $ObjectType=null){
		if($Enable)
			$this->RegisterObjectMessages($LocaleID, $ObjectType);
		else 
			$this->UnRegisterObjectMessages($LocaleID, $ObjectType);
	}

	
}
class BaseModuleShares extends BaseModuleMessages {
	
	public function Create(){
		parent::Create();
		$this->RegisterPropertyString('ShareList', "");
		$this->RegisterTimer('FlushTimer',0,"IPS_RequestAction($this->InstanceID,'FLUSH_TIMER',true);");
		$this->SetBuffer('FlushShares', '');
		$this->SetBuffer('ShareIndex', '');
		if($this->ConnectParentGuid)$this->ConnectParent($this->ConnectParentGuid);
	}
	public function RequestAction($Ident, $Value){
		if($Ident=='FLUSH_TIMER'){
			$this->FlushShareCache();
		}
		else parent::RequestAction($Ident, $Value);
	}
	// -----------------------------------------
	// - Basic Maintain
	// -----------------------------------------
	protected function isValidSyncMode($SyncMode){
	
		return $SyncMode == DIR_ALL || $SyncMode == $this->IsServer?DIR_STOC:DIR_CTOS;
	}
	protected function ApplyCleanUp(){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug( "execute",__FUNCTION__);
		$this->SetTimerInterval('FlushTimer',0);
		$index=$this->LoadShareIndex();
		if(count($index)>0){
			
			foreach($index as $share){
				$this->OnShareDelete($share);
			}
			$this->SaveShareIndex([]);
		}	
		$this->SetBuffer('BlockFlushShareCache','');
		$this->SetBuffer('FlushShares', '');
		$this->SetBuffer('FlushEnabled','');
	}
	
	protected function OnFlushTimer(){}
	// -----------------------------------------
	// -   Data Events
	// -----------------------------------------
	protected function OnRequestReceived($RequestType, $Request,int $Id, string $ClientIP, int $ClientPort){
		parent::OnRequestReceived($RequestType, $Request, $Id, $ClientIP, $ClientPort);
		if($RequestType==RTYPE_SHARES ){
			$this->OnSharesRequest($Request, $Id, $ClientIP, $ClientPort);
		}
		elseif($RequestType==RTYPE_CHANGES ){
			$this->OnSharesChangesRequest($Request, $Id, $ClientIP, $ClientPort);
		}
	}
	protected function OnResponseReceived($ResponseType, $Response, bool $Ack, int $Id, string $ClientIP, int $ClientPort){
		parent::OnResponseReceived($ResponseType, $Response, $Ack, $Id, $ClientIP, $ClientPort);
		if($ResponseType==RTYPE_SHARES ){
			$this->OnSharesResponse($Response[0],(bool)$Response[1], $Ack,$Id, $ClientIP, $ClientPort);
		}
		elseif($ResponseType==RTYPE_CHANGES ){
			$this->OnSharesChangesResponse($Response, $Ack, $Id, $ClientIP, $ClientPort);
		}
	}
	protected function OnSharesResponse(array $Shares,bool $Filtered,bool $Ack, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?"from $ClientIP:$ClientPort ":'';
			$this->Debug( $m."ID: $Id Filtered: ".$this->Bool2Str($Filtered,BOOL_YESNO)." Ack: ".$this->Bool2Str($Ack,2).' => '.json_encode($Shares),__FUNCTION__);
		}
	}
	protected function OnSharesRequest(array $FilterObjectIDs, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?"from $ClientIP:$ClientPort ":'';
			$this->Debug($m."ID: $Id ObjectIDs => ".json_encode($FilterObjectIDs),__FUNCTION__);
		}
	}
	protected function OnSharesChangesResponse(array $ShareChangesResponse,bool $Ack, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?"to $ClientIP:$ClientPort ":'';
			$this->Debug($m."ID: $Id Ack: ".$this->Bool2Str($Ack,2).' => '.json_encode($ShareChangesResponse),__FUNCTION__);
		}
		if($this->HasDiffChanges($ShareChangesResponse)){
			$this->ApplyShareChangesResponse($ShareChangesResponse);
		}
	}
	protected function OnSharesChangesRequest(array $ShareChangesRequest, int $Id, string $ClientIP, int $ClientPort){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?"from $ClientIP:$ClientPort ":'';
			$this->Debug($m."ID: $Id ObjectIDs => ".json_encode($ShareChangesRequest),__FUNCTION__);
		}
		$this->DoSendChangesResponse($this->GetShareChangesResponse($ShareChangesRequest),$ClientIP, $ClientPort,$Id);
	}
	protected function OnBeforeShareListSaved(array &$Shares, array $OldShares){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(json_encode($Shares),__FUNCTION__);
		
	}
	protected function OnShareListSaved(array $Shares){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(json_encode($Shares),__FUNCTION__);
		
	}
	protected function OnShareIndexSaved(array $SharesIndex){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(json_encode($SharesIndex),__FUNCTION__);
		
	}
	protected function OnShareAdd(array &$Share){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug(json_encode($Share),__FUNCTION__);
	}
	protected function OnShareDelete(array $Share){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug(json_encode($Share),__FUNCTION__);
	}
	protected function OnShareChanged(array $Share, $OldShare){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$this->Debug('old: '.json_encode($OldShare),__FUNCTION__);
			$this->Debug('new: '.json_encode($Share),__FUNCTION__);
		}
		elseif(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug(json_encode($Share),__FUNCTION__);
		if($OldShare[SHARE_STATE]!=$Share[SHARE_STATE]  || $OldShare[SHARE_DIR]!=$Share[SHARE_DIR]){
			$this->OnShareStateChanged($Share,$OldShare[SHARE_STATE], $OldShare[SHARE_DIR] );
 		}
	}
	protected function OnShareStateChanged($Share, int $OldState, int $OldDirection){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("OldState: $OldState OldDir: $OldDirection => ".json_encode($Share),__FUNCTION__);
		
		if($Share[SHARE_STATE]==STATE_CONNECTED){
			$this->OnShareEnabled($Share);
		}elseif($OldState==STATE_CONNECTED){
			$this->OnShareDisabled($Share);
		}
	}
	protected function OnShareEnabled(array $Share){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug(json_encode($Share),__FUNCTION__);
	}
	protected function OnShareDisabled(array $Share){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug(json_encode($Share),__FUNCTION__);
	}
	protected function OnShareChange(array $Share, $ValueType, $OldValue, $NewValue){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			if(is_string($OldValue)){
				if(strlen($OldValue)>20)$OldValue=substr($OldValue,0,20).'...';
			}elseif(is_bool($OldValue))$OldValue=$this->Bool2Str($OldValue,0);
			if(is_string($NewValue)){
				if(strlen($NewValue)>20)$NewValue=substr($NewValue,0,20).'...';
			}elseif(is_bool($NewValue))$NewValue=$this->Bool2Str($NewValue,0);
			$this->Debug("ValueType: $ValueType Value Old: $OldValue New: $NewValue => ". json_encode($Share),__FUNCTION__);
		}
		
	}
	// -----------------------------------------
	// -   Data Send Maintain
	// -----------------------------------------
	protected function DoSendShareRequest(array $FilterObjectIDs, $ClientIP='',$ClientPort=0,$Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(json_encode($FilterObjectIDs),__FUNCTION__);
		$this->DoSendRequest(RTYPE_SHARES,$FilterObjectIDs, $ClientIP,$ClientPort,$Id);
	}
	protected function DoSendShareResponse(array $Shares,bool $Filtered,$Ack=false, $ClientIP='',$ClientPort=0, $Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Filtered: ".$this->Bool2Str($Filtered).' => '.json_encode($Shares),__FUNCTION__);
		$this->DoSendResponse(RTYPE_SHARES,[$Shares,$Filtered],$Ack, $ClientIP,$ClientPort,$Id);
	}
	protected function DoSendChangesRequest(array $ShareChangesRequest, $ClientIP='',$ClientPort=0,$Id=0){
 		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(json_encode($ShareChangesRequest),__FUNCTION__);
		$this->DoSendRequest(RTYPE_CHANGES,$ShareChangesRequest, $ClientIP,$ClientPort,$Id);
	}
	protected function DoSendChangesResponse(array $ShareChangesResponse, $Ack=false,$ClientIP='',$ClientPort=0,$Id=0){
 		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(json_encode($ShareChangesResponse),__FUNCTION__);
		$this->DoSendResponse(RTYPE_CHANGES,$ShareChangesResponse, $Ack, $ClientIP,$ClientPort,$Id);
	}
	
	// -----------------------------------------
	// -   Shares Maintain
	// -----------------------------------------
	private $shareCache=null;	
	protected function LoadShareList($Silent=false, $IgnoreFlushCache=false){
		if(!$IgnoreFlushCache && ($shares=json_decode($this->GetBuffer('FlushShares'),true))){
			if(!$Silent){
				if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Return from Flush-cache => ".json_encode($shares),__FUNCTION__);
				elseif(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Return from Flush-cache",__FUNCTION__);
			}
			return $this->shareCache=$shares;
		}
		if(!is_null($this->shareCache)){
			if(!$Silent){
				if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Return from cache => ".json_encode($this->shareCache),__FUNCTION__);
				elseif(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Return from cache",__FUNCTION__);
			}
			return $this->shareCache;
		}
		$this->shareCache= ($l=json_decode($this->ReadPropertyString('ShareList'),true))?$l:[];
		if(!$Silent && REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug(json_encode($this->shareCache),__FUNCTION__);				
		return $this->shareCache;
	}
	protected function SaveShareList(array $Shares, $DoEvent=true, $DoCache=false){
		if(json_encode($Shares)!=json_encode($OldShares=$this->LoadShareList())){
			$this->shareCache=$Shares;
			$encoded=json_encode($Shares);
			if(!$DoCache && empty($this->GetBuffer('FlushEnabled'))){
				if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Event: ".$this->Bool2Str($DoEvent,BOOL_YESNO)." Cache:".$this->Bool2Str($DoCache,BOOL_YESNO)." => $encoded",__FUNCTION__);		
				elseif(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Saved",__FUNCTION__);		
				$this->OnBeforeShareListSaved($Shares, $OldShares);
				IPS_SetProperty($this->InstanceID, 'ShareList', $encoded);
				IPS_ApplyChanges($this->InstanceID);
				if($DoEvent)$this->OnShareListSaved($Shares);
			} else $this->PushShareCache($encoded);
 			return true;
		}
		return false;
	}
	protected function SendShares($Shares=null,array $FilterObjectIDs=null, bool $Ack=false, string $ClientIP='',int $ClientPort=0, int $Id=0){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug($m."ID: $Id Filter => ".json_encode($FilterObjectIDs),__FUNCTION__);		
		}
		if(is_null($Shares))$Shares=$this->LoadShareList();
		if($filtered=is_array($FilterObjectIDs)&&  count($FilterObjectIDs)>0){
			$Shares=array_filter($Shares,function($i)use($FilterObjectIDs){return in_array($i[SHARE_OBJ_ID], $FilterObjectIDs);});
			$Shares=array_values($Shares);
		}
		foreach($Shares as &$share){
 			if($filtered){
				unset($share['id']);
				unset($share['parent']);
			}
			if($this->IsServer)$share[SHARE_NAME]=IPS_GetName($share[SHARE_OBJ_ID]);else unset($share[SHARE_LOCALID]);
 		}
		$this->DoSendShareResponse($Shares,$filtered, $Ack, $ClientIP,$ClientPort,$Id);
	}
	protected function ReSortSharesByParent(&$Shares){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("exeucute", __FUNCTION__);
		$newIndex=1;
		$roots = array_filter($Shares,function($s){return $s[SHARE_PARENT]==0;});
		foreach($roots as $index=>$share){
			$Shares[$index]['id']=$parentID=$newIndex++;
			unset($Shares[$index]['parent']);
			foreach($Shares as &$subshare){
				if($subshare[SHARE_PARENT]==$share[SHARE_OBJ_ID]){
					$subshare['id']=$newIndex++;
					$subshare['parent']=$parentID;
				}
			}
		}
	}
	protected function GetShare(int $ObjectID){
		$shareList	= $this->LoadShareList(true);
		$found		= array_filter($shareList,function($c)use($ObjectID){return $c[SHARE_OBJ_ID]==$ObjectID;});
		$index		= key($found);
		return 	is_null($index)?null:$found[$index];
	}
	protected function PutShare(array $Share, $CacheChanges=true){
		$shareList	= $this->LoadShareList(true);
		$found		= array_filter($shareList,function($c)use($Share){return $c[SHARE_OBJ_ID]==$Share[SHARE_OBJ_ID];});
		$index		= key($found);
		$changed	= false;
		if(!is_null($index)){
			foreach($Share as $valueType=>$value){
				if($shareList[$index][$valueType]!=$value)$changed=true;
				$shareList[$index][$valueType]=$value;
			}
		}
		if($changed){
			$this->SaveShareList($shareList,false,$CacheChanges);
		}
		return $changed;
	}
	protected function GetShareChangesResponse(array $ShareChangesRequest){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Request: ".json_encode($ShareChangesRequest),__FUNCTION__);		
		$shareIndex=$this->LoadShareIndex();
		$myKeys	= array_keys($shareIndex);
		$reqKeys= array_keys($ShareChangesRequest);
		$delete	= array_diff($reqKeys,$myKeys);
		$add	= array_diff($myKeys,$reqKeys);
		$diff	= [DIFF_UPDATE=>[],DIFF_DELETE=>[],DIFF_ADD=>[] ];
		$clean  = function($share, $IncludeName){
			unset($share['id']);
			unset($share['parent']);
			unset($share[SHARE_LOCALID]);
			unset($share[SHARE_STATE]);
			if($IncludeName)$share[SHARE_NAME]=IPS_GetName($share[SHARE_OBJ_ID]);
			return $share;
		};
		foreach ($delete as $objectID){
			$diff[DIFF_DELETE][$objectID]=1;
			unset($shareIndex[$objectID]);
		}
		foreach ($add as $objectID){
			$diff[DIFF_ADD][$objectID]=$clean($shareIndex[$objectID],$this->IsServer);
			unset($shareIndex[$objectID]);
		}
		foreach ($shareIndex as $objectID=>$share){
			foreach($share as $valueType=>$value){
				if(in_array($valueType, [SHARE_DIR]) &&
					$ShareChangesRequest[$objectID][$valueType]!=$value){
					$diff[DIFF_UPDATE][$objectID][$valueType]=$value;
				}
			}
		}
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Response:".json_encode($diff),__FUNCTION__);		
		return $diff;
	}
	protected function ApplyShareChangesResponse(array $ShareChangesResponse){
		$shareIndex = $this->LoadShareIndex();
		$shareList	= $this->LoadShareList();
		$sharesKeys = array_map(function($s){return $s[SHARE_OBJ_ID];},$shareList);
		$reSort=false;
		foreach (array_keys($ShareChangesResponse[DIFF_DELETE]) as $objectID){
			$index=array_search($objectID, $sharesKeys);
			if($index!==false){
				$this->OnShareDelete($shareList[$index]);
				unset($shareList[$index]);
				$reSort=true;
			}
			unset($shareIndex[$objectID]);
		}
		foreach ($ShareChangesResponse[DIFF_ADD] as $objectID=>$share){
			$newShare=[SHARE_OBJ_ID=>0,SHARE_DIR=>0,SHARE_TYPE=>0,SHARE_STATE=>0,SHARE_PARENT=>0,SHARE_VTYPE=>0,SHARE_NAME=>'',SHARE_LOCALID=>0];
			foreach($share as $valueType=>$value)$newShare[$valueType]=$value;
			$this->OnShareAdd($newShare);
			$shareList[]=$newShare;
			$shareIndex[$objectID]=$newShare;
			$reSort=true;
		}
		foreach ($ShareChangesResponse[DIFF_UPDATE] as $objectID=>$share){
			$index=array_search($objectID, $sharesKeys);
			$changed = false;
			if($index!==false){
				$oldShare=$shareList[$index];
				foreach($share as $valueType=>$value){
					if(!in_array($valueType, [SHARE_STATE,SHARE_LOCALID]) &&  $shareList[$index][$valueType]!=$value){
						$this->OnShareChange($shareList[$index], $valueType, $shareList[$index][$valueType], $value);
						$shareList[$index][$valueType]=$value;
						$changed=true;
					}
				}
				$shareIndex[$objectID]=$shareList[$index];
				if($changed)$this->OnShareChanged($shareList[$index], $oldShare);
			}
		}
		if($reSort)$this->ReSortSharesByParent($shareList);
		$this->SaveShareIndex($shareIndex);
		$this->SaveShareList($shareList);
	}
	private function FlushShareCache(){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("exeucute", __FUNCTION__);
		$this->SetTimerInterval('FlushTimer', 0);
		$this->SetBuffer('BlockFlushShareCache',1);
		$this->OnFlushTimer();
		$this->SetBuffer('BlockFlushShareCache','');
		$shares=$this->GetBuffer('FlushShares');
		if(!empty($shares)){
			$this->SetBuffer('FlushShares', '');
			$shares=json_decode($shares,true);
			$this->OnBeforeShareListSaved($shares, $this->LoadShareList());
			$this->shareCache=$shares;
			IPS_SetProperty($this->InstanceID, 'ShareList', json_encode($this->shareCache));
		}
		$this->SetBuffer('FlushEnabled','');
		if(IPS_HasChanges($this->InstanceID))IPS_ApplyChanges($this->InstanceID);
		if($shares)$this->OnShareListSaved($this->shareCache);
	}
	private function PushShareCache(string $JSONShares){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug($JSONShares,__FUNCTION__);
		$this->SetBuffer('FlushShares', $JSONShares);
		$this->SetBuffer('FlushEnabled',empty($JSONShares)?'':1);
		if(empty($this->GetBuffer('BlockFlushShareCache')))$this->SetTimerInterval('FlushTimer', REMOTEIO_FLUSH_TIMEOUT);
	}
	// -----------------------------------------
	// -   Shares Index Maintain
	// -----------------------------------------
	private $indexCache=null;
	protected function CreateShareIndex(){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("execute",__FUNCTION__);
		$objectIDs=array_map(function($s){return $s[SHARE_OBJ_ID];},$shares=$this->LoadShareList());
		$index=array_combine($objectIDs,$shares);
		$this->SetBuffer('ShareIndex',json_encode($index));
		foreach($index as $objectID=>$share){
			if($this->IsServer){
				$this->RegisterObject($objectID,$share[SHARE_TYPE]);
				if($share[SHARE_STATE]==STATE_CONNECTED){
					$this->OnShareEnabled($share);
				}
			}elseif($share[SHARE_LOCALID]){
				$this->RegisterObject($share[SHARE_LOCALID],$share[SHARE_TYPE]);
				$this->OnShareEnabled($share);
			}
		}
		return $this->indexCache=$index;
	}
	protected function LoadShareIndex($Silent=false){
		if(!is_null($this->indexCache)){
			if(!$Silent){
				if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("Return from cache => ".json_encode($this->indexCache),__FUNCTION__);
				elseif(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Return from cache",__FUNCTION__);
			}
			return $this->indexCache;
		}
		$this->indexCache=!is_null($l=json_decode($this->GetBuffer('ShareIndex'),true))?$l:$this->CreateShareIndex(); 
		if(!$Silent && REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug(json_encode($this->indexCache),__FUNCTION__);
		return $this->indexCache;
	}
	protected function SaveShareIndex(array $ShareIndex, $DoEvent=true){
		$encoded=json_encode($ShareIndex);
		if($encoded!=json_encode($this->LoadShareIndex(true))){
			$this->indexCache=$ShareIndex;
			if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Event: ".$this->Bool2Str($DoEvent,BOOL_YESNO)." => $encoded", __FUNCTION__);		
			elseif(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug("Saved",__FUNCTION__);		
			$this->SetBuffer('ShareIndex',$encoded);
			if($DoEvent)$this->OnShareIndexSaved($ShareIndex);
		}
	}
	protected function ShareIndexDiffToShareChangesResponse(array $ShareIndexDiff){
		$clean  = function($share, $IncludeName){
			unset($share['id']);
			unset($share['parent']);
			unset($share[SHARE_LOCALID]);
			unset($share[SHARE_STATE]);	
			if($IncludeName)$share[SHARE_NAME]=IPS_GetName($share[SHARE_OBJ_ID]);
			return $share;
		};
		foreach ($ShareIndexDiff[DIFF_DELETE] as $objectID=>$share){
			$ShareIndexDiff[DIFF_DELETE][$objectID]=true;
		}
		foreach($ShareIndexDiff[DIFF_ADD] as $objectID=>$share){
			$ShareIndexDiff[DIFF_ADD][$objectID]=$clean($share,true);
		}
		foreach($ShareIndexDiff[DIFF_UPDATE] as $objectID=>$share){
			$share=$clean($share,false);
			if(count($share)>0)
				$ShareIndexDiff[DIFF_UPDATE][$objectID]=$share;
			else 
				unset($ShareIndexDiff[DIFF_UPDATE][$objectID]);
		}
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug(json_encode($ShareIndexDiff),__FUNCTION__);	
		return $ShareIndexDiff;
	}
	protected function GetSharesIndexDiff(array $MyShares, array $ShareIndex){
		if(is_null($MyShares))$MyShares=$this->LoadShareList();
		if(is_null($ShareIndex))$ShareIndex=$this->LoadShareIndex();
		$myKeys	= array_map(function($s){return $s[SHARE_OBJ_ID];},$MyShares);
		$newKeys= array_keys($ShareIndex);
		$delete	= array_diff($newKeys,$myKeys);
		$add	= array_diff($myKeys,$newKeys);
		$diff	= [DIFF_UPDATE=>[],DIFF_DELETE=>[],DIFF_ADD=>[] ];
		foreach ($delete as $objectID)$diff[DIFF_DELETE][$objectID]=$ShareIndex[$objectID];
		foreach($add as $objectID)$diff[DIFF_ADD][$objectID]=$MyShares[array_search($objectID, $myKeys)];
		foreach($MyShares as $share){
			$objectID=$share[SHARE_OBJ_ID];
			if(in_array($objectID, $add))continue;
			foreach($share as $valueType=>$value){
				if(in_array($valueType, [SHARE_DIR,SHARE_STATE]) && $ShareIndex[$objectID][$valueType]!=$value)
					$diff[DIFF_UPDATE][$objectID][$valueType]=$value;
			}
		}
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug(json_encode($diff),__FUNCTION__);		
		return $diff;
	}
	protected function ApplySharesIndexDiff(array $ShareIndexDiff, array &$ShareIndex=null){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug(json_encode($ShareIndexDiff),__FUNCTION__);		
		if(is_null($ShareIndex))$MyShareIndex=$this->LoadShareIndex();else $MyShareIndex=&$ShareIndex;
		foreach ($ShareIndexDiff[DIFF_DELETE] as $objectID=>$share){
			$this->OnShareDelete($share);
			unset($MyShareIndex[$objectID]);
		}
		foreach($ShareIndexDiff[DIFF_ADD] as $objectID=>$share){
			$MyShareIndex[$objectID]=$share;
			$this->OnShareAdd($MyShareIndex[$objectID]);
		}
		foreach($ShareIndexDiff[DIFF_UPDATE] as $objectID=>$share){
			$changed=false;$oldShare=$MyShareIndex[$objectID];
			foreach($share as $valueType=>$value){
				if(in_array($valueType, [SHARE_DIR,SHARE_STATE]) &&
						$oldShare[$valueType]!=$value){
					$this->OnShareChange($MyShareIndex[$objectID],$valueType,$oldShare[$valueType],$value);
					$MyShareIndex[$objectID][$valueType]=$value;
					$changed=true;
				}
			}
			if($changed)$this->OnShareChanged($MyShareIndex[$objectID],$oldShare);
		}
		
		if(is_null($ShareIndex))$this->SaveShareIndex($MyShareIndex);
	}
	
	// -----------------------------------------
	// -   Misc
	// -----------------------------------------
	protected function HasDiffChanges(array $Diff){
		$r=!empty($Diff[DIFF_UPDATE])||!empty($Diff[DIFF_DELETE])||!empty($Diff[DIFF_ADD]);
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Returns => ".$this->Bool2Str($r),__FUNCTION__);
		return $r;
	}
	protected function HasDiffUpdatesOnly(array $Diff){
		$r=!empty($Diff[DIFF_UPDATE]) && empty($Diff[DIFF_DELETE]) && empty($Diff[DIFF_ADD]);
		if(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("Returns => ".$this->Bool2Str($r),__FUNCTION__);
		return $r;
		
	}
	
	
}

class BaseModuleCommand extends BaseModuleShares {
	protected $DefaultCommands = [CMD_SHARES_CHANGED];
	public function Create(){
		parent::Create();
		$this->RegisterPropertyInteger('AllowRemoteCommands', COMMAND_ALLOW_NONE);
	}
	
	// -----------------------------------------
	// -   Command Events
	// -----------------------------------------
	protected function OnRequestReceived($RequestType, $Request,int $Id, string $ClientIP, int $ClientPort){
		parent::OnRequestReceived($RequestType, $Request, $Id, $ClientIP, $ClientPort);
		if($RequestType==RTYPE_COMMAND ){
			if(in_array($Request[1] , $this->DefaultCommands)){
				if($Request[1]==CMD_SHARES_CHANGED){
					$this->OnCommandSharesChanged($Request[2]);
				}else $this->OnDefaultCommandRequest($Request[0], $Request[1], $Request[2], $ClientIP, $ClientPort, $Id);
			}
			elseif($this->ReadPropertyInteger('AllowRemoteCommands')>COMMAND_ALLOW_NONE){
				$this->OnCommandRequest($Request[0],$Request[1],$Request[2], $ClientIP, $ClientPort,$Id);
			}
			else{
				$m=$ClientIP?" from $ClientIP:$ClientPort":'';
				if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Identifier: {$Request[0]} Command: {$Request[1]} ID: $Id$m !! Remote commands not allowed on this server" ,__FUNCTION__);
				IPS_LogMessage(IPS_GetName($this->InstanceID),"ERROR! Remote commands not allowed on this server! Identifier: {$Request[0]} Command: {$Request[1]} ID: $Id$m");
			}
		}
		
	}
	protected function OnResponseReceived($ResponseType, $Response, bool $Ack, int $Id, string $ClientIP, int $ClientPort){
		parent::OnResponseReceived($ResponseType, $Response, $Ack, $Id, $ClientIP, $ClientPort);
		if($ResponseType==RTYPE_COMMAND){
			if(in_array($Response[1] , $this->DefaultCommands)){
				
				$this->OnDefaultCommandResponse($Response[0],$Response[1],$Response[2], $Ack, $ClientIP, $ClientPort,$Id);
			}
			elseif($this->ReadPropertyInteger('AllowRemoteCommands')>COMMAND_ALLOW_NONE){
				$this->OnCommandResponse($Response[0],$Response[1],$Response[2], $Ack, $ClientIP, $ClientPort,$Id);
			}
			else{
				$m=$ClientIP?" from $ClientIP:$ClientPort":'';
				if(REMOTEIO_DEBUG>DEBUG_NONE)$this->Debug("Command: {$Response[0]} ID: $Id$m !! Remote commands not allowed on this server" ,__FUNCTION__);
				IPS_LogMessage(IPS_GetName($this->InstanceID),"ERROR! Remote commands not allowed on this server! Command: {$Response[0]} ID: $Id$m");
			}
			
		}

	}
	// -----------------------------------------
	// -  Command Internal
	// -----------------------------------------
	protected function OnCommandSharesChanged($Arguments){
		if(REMOTEIO_DEBUG>DEBUG_ERROR)$this->Debug(json_encode($Arguments),__FUNCTION__);
	}
	
	// -----------------------------------------
	// -  Command Request/Response
	// -----------------------------------------
	protected function OnCommandRequest(string $Identifier, string $Command,array $Arguments,string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $Identifier Command: $Command ID: $Id$m Args => ".json_encode($Arguments),__FUNCTION__);
		}
	}
	protected function OnCommandResponse(string $Identifier, string $Command, $CommandResult,  bool $Ack, string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $Identifier Command: $Command Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($CommandResult),__FUNCTION__);
		}
	}
	protected function OnDefaultCommandRequest(string $Identifier, string $Command,array $Arguments,string $ClientIP,int $ClientPort ,int $Id ){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Identifier: $Identifier Command: $Command ID: $Id$m Args => ".json_encode($Arguments) ,__FUNCTION__);
		}
		
	}
	protected function OnDefaultCommandResponse(string $Identifier, string $Command,$CommandResult,bool $Ack,string $ClientIP,int $ClientPort ,int $Id ){
		if(REMOTEIO_DEBUG>DEBUG_ERROR){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Identifier: $Identifier Command: $Command Ack: ".$this->Bool2Str($Ack)." ID: $Id$m Returns => ".json_encode($CommandResult) ,__FUNCTION__);
		}
		
	}
	
	// -----------------------------------------
	// -  Command Sending
	// -----------------------------------------
	protected function DoSendCommandRequest(string $Command,array $Arguments=[],$ClientIP='',$ClientPort=0 ,$Id=0){
		$ident=$this->ReadPropertyString('Identifier');
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Command: $Command ID: $Id$m => ".json_encode($Arguments),__FUNCTION__);
		}
		$this->DoSendRequest(RTYPE_COMMAND,[ $ident,$Command,$Arguments], $ClientIP,$ClientPort,$Id);
	}
	protected function DoSendCommandResponse(string $Command,$CommandResult, $Ack=false, $ClientIP='',$ClientPort=0 ,$Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Command: $Command Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($CommandResult),__FUNCTION__);
		}
		$this->DoSendResponse(RTYPE_COMMAND, [$Command,$CommandResult], $Ack, $ClientIP,$ClientPort,$Id);
	}
	protected function SendCommandSharesChanged( $ClientIP='',$ClientPort=0 ,$Id=0){
		$this->DoSendCommandRequest(CMD_SHARES_CHANGED,[],$ClientIP,$ClientPort,$Id);
	}
	// -----------------------------------------
	// -  Command Execute
	// -----------------------------------------
	protected function CommandExecute(string $Command, array $Arguments){
		
	}
}

class BaseModuleConfig extends BaseModuleCommand {
	// -----------------------------------------
	// -   config Events
	// -----------------------------------------
	protected function OnRequestReceived($RequestType, $Request,int $Id, string $ClientIP, int $ClientPort){
		parent::OnRequestReceived($RequestType, $Request, $Id, $ClientIP, $ClientPort);
		if($RequestType==RTYPE_CONFIG ){
			$this->OnConfigRequest($Request, $ClientIP, $ClientPort,$Id);
		}
		elseif($RequestType==RTYPE_STATE ){
			$this->OnStateRequest($Request, $ClientIP, $ClientPort,$Id);
		}
	}
	protected function OnResponseReceived($ResponseType, $Response, bool $Ack, int $Id, string $ClientIP, int $ClientPort){
		parent::OnResponseReceived($ResponseType, $Response, $Ack, $Id, $ClientIP, $ClientPort);
		if($ResponseType==RTYPE_CONFIG ){
			$this->OnConfigResponse($Response, $Ack, $ClientIP, $ClientPort,$Id);
		}
		elseif($ResponseType==RTYPE_STATE ){
			$this->OnStateResponse($Response[0],$Response[1], $Ack, $ClientIP, $ClientPort,$Id);
		}
	}
	// -----------------------------------------
	// -  config Data Request/Response
	// -----------------------------------------
	protected function OnConfigRequest(string $Ident,string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $Ident ID: $Id$m",__FUNCTION__);
		}
	}
	protected function OnConfigResponse(string $JSONConfig, bool $Ack, string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => $JSONConfig",__FUNCTION__);
		}
	}
	protected function OnStateRequest(string $Ident,string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $Ident ID: $Id$m",__FUNCTION__);
		}
		
	}
	protected function OnStateResponse(string $Ident, int $ClientState,bool $Ack, string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $Ident Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".$this->TranslatedClientState($ClientState),__FUNCTION__);
		}
		
	}
	// -----------------------------------------
	// -  config Data Sending
	// -----------------------------------------
	protected function DoSendConfigRequest($ClientIP='',$ClientPort=0 ,$Id=0){
		$ident=$this->ReadPropertyString('Identifier');
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $ident ID: $Id$m",__FUNCTION__);
		}
		$this->DoSendRequest(RTYPE_CONFIG, $ident, $ClientIP,$ClientPort,$Id);
	}
	protected function DoSendConfigResponse(string $JSONConfig, $Ack=false, $ClientIP='',$ClientPort=0 ,$Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => $JSONConfig",__FUNCTION__);
		}
		$this->DoSendResponse(RTYPE_CONFIG, $JSONConfig, $Ack, $ClientIP,$ClientPort,$Id);
	}

	protected function DoSendStateRequest($ClientIP='',$ClientPort=0 ,$Id=0){
		$ident=$this->ReadPropertyString('Identifier');
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $ident ID: $Id$m",__FUNCTION__);
		}
		$this->DoSendRequest(RTYPE_STATE, $ident, $ClientIP,$ClientPort,$Id);
	}
	protected function DoSendStateResponse(int $ClientState, $Ack=false, $ClientIP='',$ClientPort=0 ,$Id=0){
		$ident=$this->ReadPropertyString('Identifier');
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $ident Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".$this->TranslatedClientState($ClientState),__FUNCTION__);
		}
		$this->DoSendResponse(RTYPE_STATE, [$ident,$ClientState], $Ack, $ClientIP,$ClientPort,$Id);
	}
	
}

class BaseModuleConnect extends BaseModuleConfig{
	// -----------------------------------------
	// -   Connect Events
	// -----------------------------------------
	protected function OnRequestReceived($RequestType, $Request,int $Id, string $ClientIP, int $ClientPort){
		parent::OnRequestReceived($RequestType, $Request, $Id, $ClientIP, $ClientPort);
		if($RequestType==RTYPE_CONNECT ){
			$this->OnConnectRequest($Request[0],$Request[1], $ClientIP, $ClientPort,$Id);
		}
		elseif($RequestType==RTYPE_LINKED ){
			$this->OnLinkedObjectIDsRequest($Request, $Id, $ClientIP, $ClientPort);
		}	
	}
	protected function OnResponseReceived($ResponseType, $Response, bool $Ack, int $Id, string $ClientIP, int $ClientPort){
		parent::OnResponseReceived($ResponseType, $Response, $Ack, $Id, $ClientIP, $ClientPort);
		if($ResponseType==RTYPE_CONNECT ){
			$this->OnConnectResponse($Response[0],$Response[1],$Response[2], $Ack, $ClientIP, $ClientPort,$Id);
		}
		elseif($ResponseType==RTYPE_LINKED ){
			$this->OnLinkedObjectIDsResponse($Response, $Ack, $Id, $ClientIP, $ClientPort);
		}
	}
	// -----------------------------------------
	// -  Connect Data Request/Response
	// -----------------------------------------
	protected function OnConnectRequest(string $Ident, array $ShareChangesResquest,string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $Ident ID: $Id$m => ".json_encode($ShareChangesResquest),__FUNCTION__);
		}
	}
	protected function OnConnectResponse(string $Ident, int $ConnectState,array $ShareChangesResponse, bool $Ack, string $ClientIP,int $ClientPort ,int $Id){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $Ident State: ".$this->TranslatedClientState($ConnectState)." Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($ShareChangesResponse),__FUNCTION__);
		}
	}
	protected function OnLinkedObjectIDsRequest(string $Ident,int $Id , string $ClientIP,int $ClientPort ){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $Ident ID: $Id$m",__FUNCTION__);
		}
	}
	protected function OnLinkedObjectIDsResponse(array $LinkedObjectIDs,bool $Ack,int $Id, string $ClientIP,int $ClientPort  ){
		if(REMOTEIO_DEBUG>DEBUG_NOTICE){
			$m=$ClientIP?" from $ClientIP:$ClientPort":'';
			$this->Debug("Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($LinkedObjectIDs),__FUNCTION__);
		}
	}
	// -----------------------------------------
	// -  Connect Data Sending
	// -----------------------------------------
	protected function DoSendConnectRequest(array $ShareIndex , $ClientIP='',$ClientPort=0 ,$Id=0){
		$ident=$this->ReadPropertyString('Identifier');
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $ident ID: $Id$m => ".json_encode($ShareIndex),__FUNCTION__);
		}
		$this->DoSendRequest(RTYPE_CONNECT, [$ident, $ShareIndex], $ClientIP,$ClientPort,$Id);
	}
	protected function DoSendConnectResponse(int $ConnectState, array $ChangesDiff, $Ack=false, $ClientIP='',$ClientPort=0 ,$Id=0){
		$ident=$this->ReadPropertyString('Identifier');
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
			$this->Debug("Ident: $ident Ack: ".$this->Bool2Str($Ack,BOOL_YESNO)." ID: $Id$m => ".json_encode($ChangesDiff),__FUNCTION__);
		}
		$this->DoSendResponse(RTYPE_CONNECT, [$ident,$ConnectState,$ChangesDiff], $Ack, $ClientIP,$ClientPort,$Id);
	}
	protected function DoSendLinkedObjectIDsRequest($ClientIP='',$ClientPort=0,$Id=0){
 		$ident=$this->ReadPropertyString('Identifier');
 		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
 			$m=$ClientIP?" to $ClientIP:$ClientPort":'';
 			$this->Debug("Ident: $ident$m",__FUNCTION__);
 		}
 		$this->DoSendRequest(RTYPE_LINKED,$ident, $ClientIP,$ClientPort,$Id);
	}
	protected function DoSendLinkedObjectIDsResponse(array $LinkedObjectIDs, $Ack=false,$ClientIP='',$ClientPort=0,$Id=0){
		if(REMOTEIO_DEBUG>DEBUG_DETAIL){
			$m=$ClientIP?"to $ClientIP:$ClientPort => ":'';
			$this->Debug($m.json_encode($LinkedObjectIDs),__FUNCTION__);
		}
		$this->DoSendResponse(RTYPE_LINKED,$LinkedObjectIDs, $Ack, $ClientIP,$ClientPort,$Id);
	}
	// -----------------------------------------
	// -  Connect Sending
	// -----------------------------------------
	
}

class BaseModuleForm extends BaseModuleConnect {
	public function Create() {
		parent::Create();
		$this->RegisterPropertyInteger('FormType', KERNEL_VERSION<5?4:5);
	}
	public function GetConfigurationForm(){
		return json_encode($this->GetForm((KERNEL_VERSION<5)?4:$this->ReadPropertyInteger('FormType')));
	}
	
	// -----------------------------------------
	// -  Form Maintain
	// -----------------------------------------
	protected function GetForm(int $FormVersion){
		$buttons=$this->GetFormButtons($FormVersion);
		if(count($buttons)>0){
			if($FormVersion>4){
				$f=['elements'=>[["type"=>"RowLayout", "items"=>$this->GetFormButtons($FormVersion)]]];
			}else 
				$f=['elements'=>$this->GetFormButtons($FormVersion)];
		}
		$f['elements']=array_merge($f['elements'] ,$this->GetFormList($FormVersion));
	
		if($a=$this->GetFormActions($FormVersion))$f['actions']=$a;
		$f['status']=$this->GetFormStatusCodes();
		if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->Debug("(int $FormVersion) => ".json_encode($f),__FUNCTION__,'',get_class($this));
		return $f;
	}
	protected function GetFormStatusCodes(){
		$stype=$this->IsServer?'Sever':'Client';
		return [
	        [ "code"=> 102, "icon"=> "active", "caption"=> "Connection ready" ],
	        [ "code"=> 104, "icon"=> "error", "caption"=> "Parent socket disconnected" ],
			[ "code"=> 200, "icon"=> "error", "caption"=> "No parent instance connected" ],
	        [ "code"=> 201, "icon"=> "error", "caption"=> "Connected instance is not a $stype Socket"],
		];		
	}
	protected function GetFormButtons(int $FormVersion){
		$items=[];
		if(($config=$this->GetFormConfigItems($FormVersion)) &&count($config)>0){
			if($FormVersion>4){
				$items[]=["type"=> "PopupButton", "caption"=> "Options","popup"=> [
					            "caption"=> "Module Options", "items"=> $config ] ];
			}else {
				$items=$config;
			}
		}
		return $items;
	}
	protected function GetFormConfigItems(int $FormVersion){
		$items=[		            
				["type"=>"Select", "name"=>"DebugLevel","caption"=>"Debug level", "options"=>[["value"=>DEBUG_NONE, "label"=>"None"],["value"=>DEBUG_ERROR, "label"=>"Errors"],["value"=>DEBUG_NOTICE, "label"=>"Notice"],["value"=>DEBUG_DETAIL, "label"=>"Detail"],["value"=>DEBUG_DEBUG, "label"=>"Debug"]]],
            	["type"=>"NumberSpinner", "name"=>"ControlInterval","caption"=>"Control Timer interval","suffix"=>"hours"],
			    ["type"=>"CheckBox", "name"=>"SendChangesOnly","caption"=>"Send changes only"],
				["type"=>"Select", "name"=>"AllowRemoteCommands", "caption"=>"Remote Commands", "options"=> [
						["value"=>COMMAND_ALLOW_NONE, "label"=>"Disabled"],
						["value"=>COMMAND_ALLOW_IPS, "label"=>"IPSymcon only"],
						["value"=>COMMAND_ALLOW_ALL, "label"=>"All"],
					]
				]		
		];
		if(KERNEL_VERSION>=5){
			$items[]=["type"=>"Select", "name"=>"FormType", "caption"=>"Form Mode", "options"=> [
					["value"=>5, "label"=>"Version 5"],
					["value"=>54, "label"=>"Version 4"],
					["value"=>FOMR_CONSOLE, "label"=>"Windows Console"],
				]
	       	];		
		}
		return $items;
	}
	protected function GetFormList(int $FormVersion){
		return [];
	}
	private $valueCache = null;
	protected function GetFormListValues(int $FormVersion){
		if(!is_null($this->valueCache))return $this->valueCache;
		$shares=$this->LoadShareList(false,true);
		$disabled=!$this->CheckStatus(); 
		$typenames=$this->TranslatedTypeNames();
		$statenames=$this->TranslatedStateNames();
		
		foreach($shares as &$share){
			$share['type']=  $typenames[$share[SHARE_TYPE]];
			$share['state']=$disabled?$statenames[3]:$statenames[$share[SHARE_STATE]];
			if($disabled)$share['rowColor']=COLOR_DISABLED;
			elseif($share[SHARE_STATE]==STATE_ERROR)$share['rowColor']=COLOR_ERROR;
			elseif($share[SHARE_DIR]==DIR_NONE)$share['rowColor']=COLOR_DIR_DONE;
			elseif($share[SHARE_STATE]==STATE_CONNECTED)$share['rowColor']=COLOR_CONNECTED;
			elseif(COLOR_READY)$share['rowColor']=COLOR_READY;
				
		}
// $this->SendDebug(__FUNCTION__.' : '.__LINE__,json_encode($shares), 0);
		
		/*DBG*/if(REMOTEIO_DEBUG>DEBUG_DETAIL)$this->SendDebug(__FUNCTION__.' : '.__LINE__, "($FormVersion) => ".json_encode($shares), 0);
		return $this->valueCache=$shares;
	}
	protected function GetFormActions(int $FormVersion){
		return null;
	}

}

class RemoteIOModule extends BaseModuleForm {
	public function ApplyChanges(){
		parent::ApplyChanges();
		if(!$this->ApplyBlocked()){
			$this->ApplyData();
		}elseif(REMOTEIO_DEBUG>DEBUG_NOTICE)$this->Debug("BlockApplyData set, skip apply");
		$this->UnBlockApply();
		
	}
	protected function ApplyBlocked(){
		return !empty($this->GetBuffer('BlockApplyData'));
	}
	protected function BlockApply(){
		$this->SetBuffer('BlockApplyData',1);
	}
	protected function UnBlockApply(){
		$this->SetBuffer('BlockApplyData',0);
	}
	Protected function ApplyData(){}
	
	protected function OnKernelReady(){
		parent::OnKernelReady();
		$this->CreateShareIndex();
		if($connectionID=IPS_GetInstance($this->InstanceID)['ConnectionID']){
			$this->MessageSink(time(),$this->InstanceID,FM_CONNECT,[$connectionID]);
		}
	}
	protected function OnInstanceReady(){
		parent::OnInstanceReady();
		$this->RestartTimer();
	}
	protected function OnInstanceError(){
		parent::OnInstanceError();
		$this->StopTimer();
	}
	protected function OnObjectRequest($ObjectIDs, $FilterOptions, $ClientIP, $ClientPort, $Id){
		parent::OnObjectRequest($ObjectIDs, $FilterOptions, $ClientIP, $ClientPort, $Id);
		$data=[];
		foreach($ObjectIDs as $objectID){
			if($localeID=$this->GetLocaleIDforRemoteID($objectID)){
				$data[$objectID]=$this->GetObjectData($localeID, $FilterOptions);
			}
		}
		$this->DoSendObjectResponse($data, $FilterOptions,false,$ClientIP, $ClientPort, $Id);
	}
	protected function OnObjectResponse($ObjectDatas, $FilterOptions,$Ack,$ClientIP,$ClientPort, $Id){
		parent::OnObjectResponse($ObjectDatas, $FilterOptions, $Ack, $ClientIP, $ClientPort,$Id);
		$profileRequests=[];
		foreach($ObjectDatas as $objectID=>$data){
			$doPush=true;
			if(isset($data[VOPT_PROFILE])){
				$profile=$data[VOPT_PROFILE];
				if($profile && !IPS_VariableProfileExists($profile)){
					unset($data[VOPT_PROFILE]);	
					if($profile[0]!='~'){
						$profileRequests[]=$objectID;
						$doPush=false;
					}
				}
			}
			if($localeID=$this->GetLocaleIDforRemoteID($objectID)){
				$share=$this->GetShare($objectID);
 				if($this->isValidSyncMode($share[SHARE_DIR])) 
					$this->UnRegisterObjectMessages($localeID,$share[SHARE_TYPE]);
				$this->SetObjectData($localeID, $data, $FilterOptions);	
				
				if($share[SHARE_STATE]==STATE_CREATE){
					if($doPush){
						$share[SHARE_STATE]=STATE_CONNECTED;
						$this->PutShare($share);
					}
				}elseif($this->isValidSyncMode($share[SHARE_DIR])){
					IPS_Sleep(200);
					$this->RegisterObjectMessages($localeID,$share[SHARE_TYPE]);
				}
			}
		}
		
		if(count($profileRequests)>0){
			$this->DoSendProfileRequest($profileRequests,$ClientIP, $ClientPort,$Id);
		}
		
	}
	protected function OnProfileRequest($ObjectIDs, $ClientIP, $ClientPort, $Id){
		parent::OnProfileRequest($ObjectIDs, $ClientIP, $ClientPort, $Id);
		$data=[];
		foreach($ObjectIDs as $objectID){
			if($localeID=$this->GetLocaleIDforRemoteID($objectID)){
				$data[$objectID]=$this->GetProfileData($localeID);
			}
		}
		$this->DoSendProfileResponse($data,false,$ClientIP, $ClientPort, $Id);
		
	}
	protected function OnProfileResponse($ProfileDatas, $Ack, $ClientIP, $ClientPort, $Id){
		parent::OnProfileResponse($ProfileDatas, $Ack, $ClientIP, $ClientPort, $Id);
		foreach($ProfileDatas as $objectID=>$data){
			if($localeID=$this->GetLocaleIDforRemoteID($objectID)){
				$this->SetProfileData($localeID, $data);
			}
		}
		
	}
	protected function OnBeforeLocalDataChanged(int $LocaleID, int $RemoteID, int $MessageID, $NewValue){
		parent::OnBeforeLocalDataChanged($LocaleID, $RemoteID, $MessageID, $NewValue);
		$this->UnregisterMessage($LocaleID, $MessageID);
		
	}
	protected function OnAfterLocalDataChanged(int $LocaleID, int $RemoteID, int $MessageID, bool $Modified){
		parent::OnAfterLocalDataChanged($LocaleID, $RemoteID, $MessageID, $Modified);	
		$share=$this->GetShare($RemoteID);
		if($this->isValidSyncMode($share[SHARE_DIR])){
			$this->RegisterObjectMessages($LocaleID,$share[SHARE_TYPE]);
		}
		
	}
	protected function OnBeforeLocalProfileChanged(int $LocaleID, string $NewProfile){
		parent::OnBeforeLocalProfileChanged($LocaleID, $NewProfile);
		$this->UnregisterMessage($LocaleID, VM_CHANGEPROFILENAME);
	}
	protected function OnAfterLocalProfileChanged(int $LocaleID, string $NewProfile, bool $Modified){ 
		parent::OnAfterLocalProfileChanged($LocaleID, $NewProfile, $Modified);
		$remoteID=$this->GetRemoteIDforLokaleID($LocaleID);
		$share=$this->GetShare($remoteID);
		if($this->isValidSyncMode($share[SHARE_DIR]))
			$this->RegisterMessage($LocaleID, VM_CHANGEPROFILENAME);
		if($share[SHARE_STATE]==STATE_CREATE){
			$share[SHARE_STATE]=STATE_CONNECTED;
			$this->PutShare($share);
		}
	}
	
}

?>